# ovirt-imageio
# Copyright (C) 2021 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Manage imageio server.

This module exposes the public names. Anything else in this package is private
and should not be used.
"""

# flake8: noqa

from .. _internal import version

from .. _internal.server import (
    DEFAULT_CONF_DIR,
    load_config,
)

from . _api import (
    Client,
    ClientError,
    Error,
    ServerError,
)

__all__ = (
    "Client",
    "ClientError",
    "DEFAULT_CONF_DIR"
    "Error",
    "ServerError",
    "load_config",
)

__version__ = version.string



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import logging
import os

from .. import errors
from .. import extent

log = logging.getLogger("backends.memory")


def open(url, mode="r", sparse=False, dirty=False, max_connections=8,
         **options):
    """
    Open a memory backend.

    Arguments:
        url: (url): ignored, for consistency with other backends.
        mode: (str): "r" for readonly, "w" for write only, "r+" for read write.
        sparse (bool): ignored, memory backend does not support sparseness.
        dirty (bool): ignored, memory backend does not support dirty extents.
        max_connections (int): maximum number of connections per backend
            allowed on this server. Limit backends's max_readers and
            max_writers.
        **options: ignored, memory backend does not have any options.
    """
    return Backend(mode=mode, max_connections=max_connections)


class Backend:
    """
    Memory backend for testing.
    """

    def __init__(self, mode="r", data=None, max_connections=8, extents=None):
        if mode not in ("r", "w", "r+"):
            raise ValueError("Unsupported mode %r" % mode)
        log.debug("Open mode=%r max_connections=%r",
                  mode, max_connections)
        self._mode = mode
        self._buf = data or bytearray()
        # TODO: Make size constant so we can build the default extents here.
        self._extents = extents or {}
        self._dirty = False
        self._position = 0
        self._closed = False
        self._max_connections = max_connections

    def clone(self):
        """
        Return a new backend sharing the same backing buffer.
        """
        return self.__class__(
            mode=self._mode,
            data=self._buf,
            max_connections=self._max_connections,
            extents=self._extents)

    @property
    def max_readers(self):
        return self._max_connections

    @property
    def max_writers(self):
        # This backend supports resize when writing or zeroing, so it cannot
        # support more than one writer concurrently.
        return 1

    # io.BaseIO interface

    def readinto(self, buf):
        self._check_closed()
        if not self.readable():
            raise IOError("Unsupproted operation: read")

        length = min(len(buf), self.size() - self._position)
        buf[:length] = self._buf[self._position:self._position + length]
        self._position += length

        return length

    def write(self, buf):
        self._check_closed()
        if not self.writable():
            raise IOError("Unsupproted operation: write")

        length = len(buf)

        extend = self._position + length - self.size()
        if extend > 0:
            self._buf.extend(extend * b"\0")

        self._buf[self._position:self._position + length] = buf
        self._position += length
        self._dirty = True

        return length

    def tell(self):
        self._check_closed()
        return self._position

    def seek(self, n, how=os.SEEK_SET):
        self._check_closed()
        if how == os.SEEK_SET:
            self._position = n
        elif how == os.SEEK_CUR:
            self._position += n
        elif how == os.SEEK_END:
            self._position = self.size() + n
        return self._position

    def flush(self):
        self._check_closed()
        self._dirty = False

    def close(self):
        log.debug("Close")
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        try:
            self.close()
        except Exception:
            # Do not hide the original error.
            if t is None:
                raise
            log.exception("Error closing")

    # Backend interface.

    def zero(self, count):
        self._check_closed()
        if not self.writable():
            raise IOError("Unsupproted operation: zero")
        return self.write(b"\0" * count)

    @property
    def block_size(self):
        return 1

    def extents(self, context="zero"):
        self._check_closed()
        # If not configured, report single data extent.
        if not self._extents and context == "zero":
            yield extent.ZeroExtent(0, self.size(), False, False)
            return

        if context not in self._extents:
            raise errors.UnsupportedOperation(
                "Backend {} does not support {} extents"
                .format(self.name, context))

        for ext in self._extents[context]:
            yield ext

    # Debugging interface

    def readable(self):
        self._check_closed()
        return self._mode in ("r", "r+")

    def writable(self):
        self._check_closed()
        return self._mode in ("w", "r+")

    @property
    def dirty(self):
        """
        Returns True if backend was modifed and needs flushing.
        """
        return self._dirty

    @property
    def sparse(self):
        return False

    @property
    def name(self):
        return "memory"

    def size(self):
        self._check_closed()
        return len(self._buf)

    def data(self):
        return self._buf[:]

    def _check_closed(self):
        if self._closed:
            # Keeping io.FileIO behaviour.
            raise ValueError("Operation on closed backend")


class ReaderFrom(Backend):

    def read_from(self, reader, length, buf):
        _copy(reader, self, length, buf)


class WriterTo(Backend):

    def write_to(self, writer, length, buf):
        _copy(self, writer, length, buf)


def _copy(reader, writer, length, buf):
    step = len(buf)
    todo = length

    while todo > step:
        reader.readinto(buf)
        writer.write(buf)
        todo -= step

    with memoryview(buf)[:todo] as view:
        reader.readinto(view)
        writer.write(view)



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import functools
import json
import logging
import os
import subprocess
import urllib.parse

from contextlib import contextmanager

from . import nbd
from . import sockutil

log = logging.getLogger("qemu_nbd")

# Allow using non system qemu-nbd, for example built from source.
QEMU_NBD = os.environ.get("QEMU_NBD", "qemu-nbd")


class Server:

    def __init__(
            self, image, fmt, sock, export_name="", read_only=False, shared=8,
            cache=None, aio=None, discard="unmap", detect_zeroes="unmap",
            bitmap=None, backing_chain=True, offset=None, size=None,
            timeout=10.0):
        """
        Initialize qemu-nbd Server.

        Arguments:
            image (str): filename to open
            fmt (str): image format (raw, qcow2, ...)
            sock (nbd.Address): socket address to listen to
            export_name (str): expose export by name
            read_only (bool): export is read-only
            shared (int): export can be shared by specified number of clients
            cache (str): cache mode (none, writeback, ...). If not specified,
                probe image and use "none" if direct I/O is supported, and
                "writeback" if not.
            aio (str): AIO mode (native or threads). If not specified, use
                "native" if cache is "none" and "threads" otherwise.
            discard (str): discard mode (ignore, unmap)
            detect_zeroes (str): Control the automatic conversion of plain zero
                writes by the OS to driver-specific optimized zero write
                commands.  Value is one of "off", "on", or "unmap".  "unmap"
                converts a zero write to an unmap operation and can only be
                used if "discard" is set to "unmap".  The default is "unmap".
            bitmap (str): export this dirty bitmap
            backing_chain (bool): when using qcow2 format, open the backing
                chain. When set to False, override the backing chain to null.
                Unallocated extents will be read as zeroes, instead of reading
                data from the backing chain. It is possible to tell if an
                extent is allocated using the extent NBD_STATE_HOLE bit.
                Using backing_chain=False requires qemu-nbd >= 5.2.0.
            offset (int): Expose a range starting at offset in raw image.
                See BlockdevOptionsRaw type in qemu source.
            size (int): Expose a range of size bytes in a raw image.
                See BlockdevOptionsRaw type in qemu source.

        See qemu-nbd(8) for more info on these options.
        """
        self.image = image
        self.fmt = fmt
        self.sock = sock
        self.export_name = export_name
        self.read_only = read_only
        self.shared = shared
        self.cache = cache
        self.aio = aio
        self.discard = discard
        self.detect_zeroes = detect_zeroes
        self.bitmap = bitmap
        self.backing_chain = backing_chain
        self.offset = offset
        self.size = size
        self.timeout = timeout
        self.proc = None

    @property
    def url(self):
        url = self.sock.url(self.export_name)
        return urllib.parse.urlparse(url)

    def start(self):
        # Implement "auto" cache mode, planned for qemu 7.
        if self.cache is None:
            self.cache = "none" if self._can_use_direct_io() else "writeback"
            log.debug("Using cache=%r", self.cache)

        if self.aio is None:
            self.aio = "native" if self.cache == "none" else "threads"
            log.debug("Using aio=%r", self.aio)

        cmd = [
            QEMU_NBD,
            "--export-name={}".format(self.export_name),
            "--persistent",
            "--shared={}".format(self.shared),
        ]

        if self.sock.transport == "unix":
            cmd.append("--socket={}".format(self.sock.path))
        elif self.sock.transport == "tcp":
            cmd.append("--bind={}".format(self.sock.host))
            cmd.append("--port={}".format(self.sock.port))
        else:
            raise RuntimeError("Unsupported transport: {}".format(self.sock))

        if self.read_only:
            cmd.append("--read-only")

        if self.cache:
            cmd.append("--cache={}".format(self.cache))

        if self.aio:
            cmd.append("--aio={}".format(self.aio))

        if self.discard:
            cmd.append("--discard={}".format(self.discard)),

        if self.detect_zeroes:
            cmd.append("--detect-zeroes={}".format(self.detect_zeroes)),

        if self.bitmap:
            cmd.append("--bitmap={}".format(self.bitmap))

        # Always using --allocation-depth simplfy everything, but we want to
        # support RHEL users that have qemu-nbd 4.2.0. Add the option only on
        # qemu-nbd >= 5.2.0, and disable backing_chain=False otherwise.
        if version() >= (5, 2, 0):
            # qemu-nbd 6.2.0 introduced a regression, returning wrong
            # base:allocation results on the second call when accessing raw
            # image.  Since raw image always reports single allocation depth
            # extent, we can safely disable it for raw images.
            # https://lists.nongnu.org/archive/html/qemu-block/2022-01/msg00292.html
            if self.fmt != "raw":
                cmd.append("--allocation-depth")
        elif self.fmt == "qcow2" and not self.backing_chain:
            raise RuntimeError(
                "backing_chain=False requires qemu-nbd >= 5.2.0")

        # Build a 'json:{...}' filename allowing control all aspects of the
        # image.

        file = {"driver": "file", "filename": self.image}

        if self.offset is not None or self.size is not None:
            # Exposing a range in a raw file.
            image = {"driver": "raw", "file": file}
            if self.offset is not None:
                image["offset"] = self.offset
            if self.size is not None:
                image["size"] = self.size

            if self.fmt == "qcow2":
                # Add a qcow2 driver on top of the raw driver. In this case we
                # cannot have any backing file so backing_chain is ignored.
                image = {"driver": "qcow2", "file": image}
        else:
            # Exposing the entire image using raw or qcow2 format.
            image = {"driver": self.fmt, "file": file}

            if self.fmt == "qcow2" and not self.backing_chain:
                image["backing"] = None

        cmd.append("json:" + json.dumps(image))

        log.debug("Starting qemu-nbd %s", cmd)
        self.proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)

        if not sockutil.wait_for_socket(self.sock, self.timeout):
            self.stop()
            raise RuntimeError("Timeout waiting for qemu-nbd socket")

        log.debug("qemu-nbd socket ready")

    def stop(self):
        if self.proc:
            log.debug("Terminating qemu-nbd gracefully")
            self.proc.terminate()

            try:
                _, err = self.proc.communicate(self.timeout)
            except subprocess.TimeoutExpired:
                log.warning("Timeout terminating qemu-nbd - killing it")
                self.proc.kill()
                _, err = self.proc.communicate(self.timeout)

            if self.proc.returncode == 0:
                log.debug("qemu-nbd terminated normally err=%r", err)
            else:
                log.warning("qemu-nbd failed rc=%s err=%r",
                            self.proc.returncode, err)

            self.proc = None

    def _can_use_direct_io(self):
        flags = os.O_RDONLY if self.read_only else os.O_RDWR
        flags |= os.O_DIRECT
        try:
            fd = os.open(self.image, flags)
        except OSError:
            return False
        else:
            os.close(fd)
            return True


@contextmanager
def run(image, fmt, sock, export_name="", read_only=False, shared=8,
        cache=None, aio=None, discard="unmap", detect_zeroes="unmap",
        bitmap=None, backing_chain=True, offset=None, size=None, timeout=10.0):
    server = Server(
        image, fmt, sock,
        export_name=export_name,
        read_only=read_only,
        shared=shared,
        cache=cache,
        aio=aio,
        discard=discard,
        detect_zeroes=detect_zeroes,
        bitmap=bitmap,
        backing_chain=backing_chain,
        offset=offset,
        size=size,
        timeout=timeout)
    server.start()
    try:
        yield
    finally:
        server.stop()


@contextmanager
def open(image, fmt, read_only=False, bitmap=None, discard="unmap",
         detect_zeroes="unmap", backing_chain=True, offset=None, size=None):
    """
    Open nbd client for accessing image using qemu-nbd.
    """
    sock = nbd.UnixAddress(image + ".sock")
    with run(
            image, fmt, sock,
            read_only=read_only,
            bitmap=bitmap,
            discard=discard,
            detect_zeroes=detect_zeroes,
            backing_chain=backing_chain,
            offset=offset,
            size=size):
        with nbd.Client(sock, dirty=bitmap is not None) as c:
            yield c


@functools.lru_cache(maxsize=1)
def version():
    """
    Return current version tuple (major, minir, patch).
    """
    # Typical output:
    # qemu-nbd 5.1.0 (qemu-kvm-5.1.0-20.el8)
    # ...
    out = subprocess.check_output([QEMU_NBD, "--version"])
    first_line = out.decode("utf-8").splitlines()[0]
    version_string = first_line.split(None, 2)[1]
    return tuple(int(n) for n in version_string.split("."))



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import nbdutil

from ovirt_imageio._internal.nbd import (
    STATE_HOLE,
    STATE_ZERO,
    EXTENT_DIRTY,
    EXTENT_BACKING,
)

MiB = 1024**2
GiB = 1024**3


def extents_length(extents):
    return sum(e.length for e in extents)


class FakeClient:

    def __init__(self, alloc, depth=None, dirty=None, max_extents=None):
        """
        alloc, depth, and dirty are list of extents of same length. The export
        size is set to the length of the extents.

        max_extents is maximum number of extents return in one extents() call
        per meta contenxt.
        """
        # Check extents total length matches.
        alloc_length = extents_length(alloc)
        if depth:
            assert extents_length(depth) == alloc_length
        if dirty:
            assert extents_length(dirty) == alloc_length

        self.export_size = alloc_length
        self.alloc = alloc
        self.depth = depth
        self.dirty = dirty
        self.max_extents = max_extents

        if self.dirty:
            self.dirty_bitmap = nbd.QEMU_DIRTY_BITMAP + "name"
        else:
            self.dirty_bitmap = None

    def extents(self, offset, length):
        """
        Simulate real NBD server extents reply.

        Return extents overlapping the requested range. The first and last
        extents may be clipped to the requested range.

        If max_extents is set, may return short reply not coverting the entire
        requested range. In this case the length of differnet meta context may
        be different.

        If export_size is shorter than the configured extents, the last extent
        may exceed the export size.
        """
        assert length > 0
        assert length <= nbd.MAX_LENGTH

        res = {
            nbd.BASE_ALLOCATION: list(
                self.lookup(offset, length, self.alloc))
        }

        if self.depth:
            res[nbd.QEMU_ALLOCATION_DEPTH] = list(
                self.lookup(offset, length, self.depth))

        if self.dirty:
            res[self.dirty_bitmap] = list(
                self.lookup(offset, length, self.dirty))

        return res

    def lookup(self, offset, length, extents):
        end = offset + length
        start = 0
        count = 0

        for e in extents:
            # Skip before the requested range:
            #   request:       [             ]
            #   extent:   |----|
            if start + e.length <= offset:
                start += e.length
                continue

            length = e.length

            # Clip extent before offset:
            #   request:       [             ]
            #   extent:    |-------|
            #   result:        |===|
            if start < offset:
                clip = offset - start
                length -= clip
                start += clip

            # Clip extent after end:
            #   request:   [             ]
            #   extent:             |-------|
            #   result:             |====|
            if start + length > end:
                clip = start + length - end
                length -= clip

            yield nbd.Extent(length, e.flags)

            # NBD server is allowed to return short reply with one or more
            # extents.
            count += 1
            if self.max_extents and count == self.max_extents:
                break

            start += length

            # Stop lookup after the requested range:
            #   request:  [             ]
            #   extent:                 |----|
            if start >= end:
                break


def fake_client(n, max_extents=0):
    """
    A client simulating few interesting cases:
    - 3 alloction types: data, zero cluster, and unallocated extent.
    - dirty extents convering both data and zero cluster.
    - extents of different meta context of different length.
    - server returning short reply.
    """
    return FakeClient(
        alloc=[
            nbd.Extent(2 * n, 0),
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        ],
        depth=[
            nbd.Extent(3 * n, 0),               # depth=1
            nbd.Extent(1 * n, 0),               # depth=2
            nbd.Extent(2 * n, EXTENT_BACKING),  # depth=0
        ],
        dirty=[
            nbd.Extent(1 * n, 0),
            nbd.Extent(3 * n, EXTENT_DIRTY),
            nbd.Extent(2 * n, 0)
        ],
        max_extents=max_extents,
    )


# Testing FakeClient


def test_fake_client_simple():
    n = MiB
    c = fake_client(n)
    res = c.extents(0, 6 * n)
    assert res == {
        nbd.BASE_ALLOCATION: c.alloc,
        nbd.QEMU_ALLOCATION_DEPTH: c.depth,
        c.dirty_bitmap: c.dirty,
    }


def test_fake_client_clip_start():
    n = MiB
    c = fake_client(n)
    res = c.extents(n, 5 * n)
    assert res == {
        nbd.BASE_ALLOCATION: [
            nbd.Extent(1 * n, 0),
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        ],
        nbd.QEMU_ALLOCATION_DEPTH: [
            nbd.Extent(2 * n, 0),               # depth=1
            nbd.Extent(1 * n, 0),               # depth=2
            nbd.Extent(2 * n, EXTENT_BACKING),  # depth=0
        ],
        c.dirty_bitmap: [
            nbd.Extent(3 * n, EXTENT_DIRTY),
            nbd.Extent(2 * n, 0)
        ],
    }


def test_fake_client_clip_end():
    n = MiB
    c = fake_client(n)
    res = c.extents(0, 5 * n)
    assert res == {
        nbd.BASE_ALLOCATION: [
            nbd.Extent(2 * n, 0),
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
            nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE),
        ],
        nbd.QEMU_ALLOCATION_DEPTH: [
            nbd.Extent(3 * n, 0),               # depth=1
            nbd.Extent(1 * n, 0),               # depth=2
            nbd.Extent(1 * n, EXTENT_BACKING),  # depth=0
        ],
        c.dirty_bitmap: [
            nbd.Extent(1 * n, 0),
            nbd.Extent(3 * n, EXTENT_DIRTY),
            nbd.Extent(1 * n, 0)
        ],
    }


def test_fake_client_clip_both():
    n = MiB
    c = fake_client(n)
    res = c.extents(2 * n,  2 * n)
    assert res == {
        nbd.BASE_ALLOCATION: [
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        ],
        nbd.QEMU_ALLOCATION_DEPTH: [
            nbd.Extent(1 * n, 0),               # depth=1
            nbd.Extent(1 * n, 0),               # depth=2
        ],
        c.dirty_bitmap: [
            nbd.Extent(2 * n, EXTENT_DIRTY),
        ],
    }


def test_fake_client_max_extents():
    n = MiB
    c = fake_client(n, max_extents=1)
    res = c.extents(0, 6 * n)
    assert res == {
        nbd.BASE_ALLOCATION: [
            nbd.Extent(2 * n, 0),
        ],
        nbd.QEMU_ALLOCATION_DEPTH: [
            nbd.Extent(3 * n, 0),               # depth=1
        ],
        c.dirty_bitmap: [
            nbd.Extent(1 * n, 0),
        ],
    }

    c = fake_client(n, max_extents=2)
    res = c.extents(n, 4 * n)
    assert res == {
        nbd.BASE_ALLOCATION: [
            nbd.Extent(1 * n, 0),
            nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        ],
        nbd.QEMU_ALLOCATION_DEPTH: [
            nbd.Extent(2 * n, 0),               # depth=1
            nbd.Extent(1 * n, 0),               # depth=2
        ],
        c.dirty_bitmap: [
            nbd.Extent(3 * n, EXTENT_DIRTY),
            nbd.Extent(1 * n, 0)
        ],
    }


# Testing nbdutil.extents()


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_all(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c))
    assert extents == [
        nbd.Extent(2 * n, 0),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE | EXTENT_BACKING),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_all_dirty(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, dirty=True))
    assert extents == [
        nbd.Extent(1 * n, 0),
        nbd.Extent(1 * n, EXTENT_DIRTY),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE | EXTENT_DIRTY),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_all_no_depth(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)

    # Simulate the case when server does not report allocation depth.
    c.depth = None

    extents = list(nbdutil.extents(c))
    assert extents == [
        nbd.Extent(2 * n, 0),
        nbd.Extent(4 * n, STATE_ZERO | STATE_HOLE),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_offset(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, offset=3 * n))
    assert extents == [
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE | EXTENT_BACKING),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_offset_dirty(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, offset=3 * n, dirty=True))
    assert extents == [
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE | EXTENT_DIRTY),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_length(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, length=3 * n))
    assert extents == [
        nbd.Extent(2 * n, 0),
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_length_dirty(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, length=3 * n, dirty=True))
    assert extents == [
        nbd.Extent(1 * n, 0),
        nbd.Extent(1 * n, EXTENT_DIRTY),
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE | EXTENT_DIRTY),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_offset_length(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, offset=n, length=4 * n))
    assert extents == [
        nbd.Extent(1 * n, 0),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE | EXTENT_BACKING),
    ]


@pytest.mark.parametrize("max_extents", [None, 1, 2])
def test_extents_offset_length_dirty(max_extents):
    n = GiB
    c = fake_client(n, max_extents=max_extents)
    extents = list(nbdutil.extents(c, offset=n, length=4 * n, dirty=True))
    assert extents == [
        nbd.Extent(1 * n, EXTENT_DIRTY),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE | EXTENT_DIRTY),
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE),
    ]


def test_extents_last_extent_excceeds_export_size():
    n = GiB
    c = fake_client(n)

    # Clip export size so we get extra extent info exceeding the request
    # length.
    c.export_size -= GiB

    # Merge base:allocation and qemu:allocation-depth.
    extents = list(nbdutil.extents(c))
    assert extents == [
        nbd.Extent(2 * n, 0),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE),
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE | EXTENT_BACKING),
    ]


def test_extents_last_extent_excceeds_export_size_dirty():
    n = GiB
    c = fake_client(n)

    # Clip export size so we get extra extent info exceeding the request
    # length.
    c.export_size -= GiB

    extents = list(nbdutil.extents(c, dirty=True))
    assert extents == [
        nbd.Extent(1 * n, 0),
        nbd.Extent(1 * n, EXTENT_DIRTY),
        nbd.Extent(2 * n, STATE_ZERO | STATE_HOLE | EXTENT_DIRTY),
        nbd.Extent(1 * n, STATE_ZERO | STATE_HOLE),
    ]


# Testing nbdutil.merged()


def test_merge_simple():
    n = GiB
    a = [nbd.Extent(n, 0)]
    b = [nbd.Extent(n, 0)]

    merged = list(nbdutil.merged(a, b))
    assert merged == a


def test_merge_split_one():
    n = GiB
    a = [
        nbd.Extent(n, 1),
        nbd.Extent(n, 2),
        nbd.Extent(n, 4),
    ]
    b = [
        nbd.Extent(n * 3, 8)
    ]

    merged1 = list(nbdutil.merged(a, b))
    assert merged1 == [
        nbd.Extent(n, 1 | 8),
        nbd.Extent(n, 2 | 8),
        nbd.Extent(n, 4 | 8),
    ]

    merged2 = list(nbdutil.merged(b, a))
    assert merged2 == merged1


def test_merge_split_both():
    n = GiB
    a = [
        nbd.Extent(n * 1, 1),
        nbd.Extent(n * 2, 2),
    ]
    b = [
        nbd.Extent(n * 2, 4),
        nbd.Extent(n * 1, 8),
    ]

    merged1 = list(nbdutil.merged(a, b))
    assert merged1 == [
        nbd.Extent(n, 1 | 4),
        nbd.Extent(n, 2 | 4),
        nbd.Extent(n, 2 | 8),
    ]

    merged2 = list(nbdutil.merged(b, a))
    assert merged2 == merged1


def test_merge_clip():
    n = GiB
    a = [
        nbd.Extent(n * 1, 1),
        nbd.Extent(n * 1, 2),
    ]
    b = [
        nbd.Extent(n * 1, 4),
        nbd.Extent(n * 2, 8),
    ]

    merged1 = list(nbdutil.merged(a, b))
    assert merged1 == [
        nbd.Extent(n * 1, 1 | 4),
        nbd.Extent(n * 1, 2 | 8),
    ]

    merged2 = list(nbdutil.merged(b, a))
    assert merged2 == merged1



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import logging

from . import errors
from . import stats
from . import util

log = logging.getLogger("ops")


class EOF(Exception):
    """ Raised when no more data is available and size was not specifed """


class Canceled(Exception):
    """ Raised when operation was canceled """


class Operation:

    # Should be overriden in sub classes.
    name = "operation"

    def __init__(self, size=None, offset=0, buf=None, clock=None):
        self._size = size
        self._offset = offset
        self._buf = buf
        self._done = 0
        self._clock = clock or stats.NullClock()
        self._canceled = False

    @property
    def size(self):
        return self._size

    @property
    def offset(self):
        return self._offset

    @property
    def done(self):
        return self._done

    @property
    def _todo(self):
        return self._size - self._done

    def run(self):
        with self._clock.run(self.name) as s:
            res = self._run()
            s.bytes += self.done
        return res

    def _run(self):
        raise NotImplementedError("Must be implemented by sub class")

    def cancel(self):
        log.debug("Cancelling operation %s", self)
        self._canceled = True

    def _record(self, name):
        """
        Return context manager for recording stats.
        """
        return self._clock.run(self.name + "." + name)

    def __repr__(self):
        return ("<{self.__class__.__name__} "
                "size={self.size} "
                "offset={self._offset} "
                "done={self.done} "
                "at 0x{id}>").format(self=self, id=id(self))


class Read(Operation):
    """
    Read data source backend to file object.
    """

    name = "read"

    def __init__(self, src, dst, buf, size, offset=0, clock=None):
        super().__init__(size=size, offset=offset, buf=buf, clock=clock)
        self._src = src
        self._dst = dst

    def _run(self):
        skip = self._offset % self._src.block_size
        self._src.seek(self._offset - skip)
        if skip:
            self._read_chunk(skip)
        while self._todo:
            self._read_chunk()

    def _read_chunk(self, skip=0):
        if self._src.tell() % self._src.block_size:
            raise errors.PartialContent(self.size, self.done)

        # If self._todo is not aligned to backend block_size we read complete
        # block and drop up to block_size - 1 bytes.
        aligned_todo = util.round_up(self._todo, self._src.block_size)

        with memoryview(self._buf)[:aligned_todo] as view:
            with self._record("read") as s:
                count = self._src.readinto(view)
                s.bytes += count
            if count == 0:
                raise errors.PartialContent(self.size, self.done)

        size = min(count - skip, self._todo)
        with memoryview(self._buf)[skip:skip + size] as view:
            with self._record("write") as s:
                self._dst.write(view)
                s.bytes += size
        self._done += size

        if self._canceled:
            raise Canceled


class Write(Operation):
    """
    Write data from file object to destination backend.
    """

    name = "write"

    def __init__(self, dst, src, buf, size=None, offset=0, flush=True,
                 clock=None):
        super().__init__(size=size, offset=offset, buf=buf, clock=clock)
        self._src = src
        self._dst = dst
        self._flush = flush

    @property
    def _todo(self):
        if self._size is None:
            return len(self._buf)
        return self._size - self._done

    def _run(self):
        try:
            self._dst.seek(self._offset)

            # If offset is not aligned to block size, receive partial chunk
            # until the start of the next block.
            unaligned = self._offset % self._dst.block_size
            if unaligned:
                count = min(self._todo, self._dst.block_size - unaligned)
                self._write_chunk(count)

            # Now current file position is aligned to block size and we can
            # receive full chunks.
            while self._todo:
                count = min(self._todo, len(self._buf))
                self._write_chunk(count)
        except EOF:
            pass

        if self._flush:
            with self._record("flush"):
                self._dst.flush()

    def _write_chunk(self, count):
        self._buf.seek(0)
        with memoryview(self._buf)[:count] as view:
            read = 0
            while read < count:
                with view[read:] as v:
                    with self._record("read") as s:
                        n = self._src.readinto(v)
                        s.bytes += n
                if not n:
                    break
                read += n

            pos = 0
            while pos < read:
                with view[pos:read] as v:
                    with self._record("write") as s:
                        n = self._dst.write(v)
                        s.bytes += n
                pos += n

        self._done += read
        if read < count:
            if self._size is None:
                raise EOF
            raise errors.PartialContent(self.size, self.done)

        if self._canceled:
            raise Canceled


class Zero(Operation):
    """
    Zero byte range.
    """

    name = "zero"

    # Limit zero size so we update self._done frequently enough to provide
    # progress even with slow storage.
    #
    # Large concurrent zero requests may lead to SCSI timeouts. These errors
    # seen on LIO server emulating thin provisioning:
    #
    #   Unable to recover from DataOut timeout while in ERL=0, closing iSCSI
    #   connection
    #
    # Limiting request size seems to avoid these timeouts. The disadvantage is
    # slower zeroing with file storage, but in this case the zeroing is
    # extremely fast so the difference is tiny.
    MAX_STEP = 128 * 1024**2

    def __init__(self, dst, size, offset=0, flush=False, clock=None):
        super().__init__(size=size, offset=offset, clock=clock)
        self._dst = dst
        self._flush = flush

    def _run(self):
        self._dst.seek(self._offset)

        while self._todo:
            step = min(self._todo, self.MAX_STEP)
            with self._record("zero") as s:
                n = self._dst.zero(step)
                s.bytes += n
            self._done += n
            if self._canceled:
                raise Canceled

        if self._flush:
            with self._record("flush"):
                self._dst.flush()


class Flush(Operation):
    """
    Flush received data to storage.
    """

    name = "flush"

    def __init__(self, dst, clock=None):
        super().__init__(clock=clock)
        self._dst = dst

    def _run(self):
        self._dst.flush()



# Copyright (C) 2021 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
import logging
import subprocess
import time

import pytest

from ovirt_imageio._internal import sockutil

from . import testutil

DAEMON_CONF = """\
[daemon]
poll_interval = 0.1
run_dir = {run_dir}
drop_privileges = no

[tls]
key_file = test/pki/system/key.pem
cert_file = test/pki/system/cert.pem
ca_file = test/pki/system/ca.pem

[remote]
port = 0

[local]
socket =

[control]
transport = {control_transport}
socket = {run_dir}/sock
port = {control_port}

[handlers]
keys = stderr

[logger_root]
level = DEBUG
handlers = stderr
"""

log = logging.getLogger("test")


@pytest.fixture(scope="module", params=["unix", "tcp"])
def srv(request, tmpdir_factory):
    tmp_dir = tmpdir_factory.mktemp("admin_tool_test")
    conf_file = tmp_dir.mkdir("conf.d").join("daemon.conf")

    random_port = testutil.random_tcp_port()

    conf = DAEMON_CONF.format(
        run_dir=str(tmp_dir),
        control_transport=request.param,
        control_port=random_port)
    conf_file.write(conf)

    log.info("Starting daemon with conf dir %s", tmp_dir)
    cmd = ["./ovirt-imageio", "--conf-dir", str(tmp_dir)]
    proc = subprocess.Popen(cmd)
    try:
        if request.param == "unix":
            socket = sockutil.UnixAddress(str(tmp_dir.join("sock")))
        else:
            socket = sockutil.TCPAddress("localhost", random_port)
        if not sockutil.wait_for_socket(socket, 10):
            raise RuntimeError("Timeout waiting for {}".format(socket))

        log.info("Daemon started with pid %s", proc.pid)
        yield str(tmp_dir)
    finally:
        log.info("Terminating daemon")
        proc.terminate()


def test_ticket_life_cycle(srv):
    log.info("Checking if ticket exits")
    with pytest.raises(subprocess.CalledProcessError) as e:
        run("show-ticket", "-c", srv, "test")
    log.debug("Error: %s", e.value.stderr)

    log.info("Adding ticket")
    run("add-ticket", "-c", srv, "test/ticket.json")

    log.info("Checking if ticket exits")
    out = run("show-ticket", "-c", srv, "test")
    ticket = json.loads(out)
    log.debug("Got ticket: %s", ticket)
    assert ticket["uuid"] == "test"

    log.info("Modifying ticket")
    modification_time = int(time.monotonic())
    run("mod-ticket", "-c", srv, "--timeout", "3000", "test")

    log.info("Checking that ticket was modifed")
    out = run("show-ticket", "-c", srv, "test")
    modified = json.loads(out)
    log.debug("Got ticket: %s", modified)
    assert modified["expires"] >= modification_time + 3000

    log.info("Removing ticket")
    run("del-ticket", "-c", srv, "test")

    log.info("Checking that ticket was removed")
    with pytest.raises(subprocess.CalledProcessError) as e:
        run("show-ticket", "-c", srv, "test")
    log.debug("Error: %s", e.value.stderr)


def run(*args):
    cmd = ["./ovirt-imageioctl"]
    cmd.extend(args)
    log.debug("Running %s", cmd)

    cp = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True)

    return cp.stdout.strip().decode("utf8")



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import io
import os

import pytest
import userstorage

from ovirt_imageio._internal import errors
from ovirt_imageio._internal import ops
from ovirt_imageio._internal import stats
from ovirt_imageio._internal import util
from ovirt_imageio._internal.backends import file
from ovirt_imageio._internal.backends import memory

from . import storage

BACKENDS = userstorage.load_config("storage.py").BACKENDS


@pytest.fixture(
    params=[
        BACKENDS["file-512-xfs"],
        BACKENDS["file-4k-xfs"],
    ],
    ids=str
)
def user_file(request):
    with storage.Backend(request.param) as backend:
        yield backend


# Common offset and size parameters.
OFFSET_SIZE = [
    pytest.param(0, 8192, id="small-aligned"),
    pytest.param(0, 511, id="small-partial-block"),
    pytest.param(42, 512 - 42 - 1, id="small-partial-block-offset"),
    pytest.param(42, 8192 - 42, id="small-unaligned-offset"),
    pytest.param(42, 8192, id="small-unaligned-offset-and-size"),
    pytest.param(0, 1024**2 * 2, id="large-aligned"),
    pytest.param(42, 1024**2 * 2 - 42, id="large-unaligned-offset"),
    pytest.param(42, 1024**2 * 2, id="large-unaligned-offset-and-size"),
]


class Operation(ops.Operation):

    def _run(self):
        if self._canceled:
            raise ops.Canceled
        self._done = 100
        return self._done


def test_run():
    op = Operation(size=100)
    assert op.offset == 0
    assert op.size == 100
    assert op.done == 0
    assert op.run() == 100
    assert op.done == 100


def test_cancel():
    op = Operation(size=100)
    assert op.offset == 0
    assert op.size == 100
    assert op.done == 0
    op.cancel()
    with pytest.raises(ops.Canceled):
        op.run()
    assert op.offset == 0
    assert op.done == 0


def test_stats(fake_time):

    class Operation(ops.Operation):

        name = "foo"

        def _run(self):

            # Simulate reading from storage...
            with self._record("read") as s:
                fake_time.now += 0.01
                s.bytes += self._size

            # Simulate writing to client socket...
            with self._record("write") as s:
                fake_time.now += 0.005
                s.bytes += self._size

            self._done = self._size

    clock = stats.Clock(now=fake_time.monotonic_time)
    op = Operation(size=8 * 1024**2, clock=clock)
    op.run()
    times = str(clock)
    assert times == (
        "[foo 1 ops, 0.015000 s, 8.00 MiB, 533.33 MiB/s] "
        "[foo.read 1 ops, 0.010000 s, 8.00 MiB, 800.00 MiB/s] "
        "[foo.write 1 ops, 0.005000 s, 8.00 MiB, 1.56 GiB/s]"
    )


@pytest.mark.parametrize("trailer", [
    pytest.param(0, id="no-trailer"),
    pytest.param(8192, id="trailer"),
])
@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_read_full(user_file, offset, size, trailer):
    data = b"b" * size

    with io.open(user_file.path, "wb") as f:
        f.write(b"a" * offset)
        f.write(data)
        f.write(b"c" * trailer)

    dst = io.BytesIO()
    with file.open(user_file.url, "r") as src, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Read(src, dst, buf, size, offset=offset)
        op.run()

    assert dst.getvalue() == data


@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_read_partial_content(user_file, offset, size):
    with io.open(user_file.path, "wb") as f:
        f.truncate(offset + size - 1)

    dst = io.BytesIO()
    with file.open(user_file.url, "r") as src, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Read(src, dst, buf, size, offset=offset)
        with pytest.raises(errors.PartialContent) as e:
            op.run()

    assert e.value.requested == size
    assert e.value.available == size - 1


def test_read_seek():
    src = memory.Backend("r", bytearray(b"0123456789"))
    src.seek(8)
    dst = io.BytesIO()
    with util.aligned_buffer(32) as buf:
        op = ops.Read(src, dst, buf, 5)
        op.run()
    assert dst.getvalue() == b"01234"


def test_read_repr():
    op = ops.Read(None, None, None, 200, offset=24)
    rep = repr(op)
    assert "Read" in rep
    assert "size=200 offset=24 done=0" in rep


@pytest.mark.parametrize("preallocated", [
    pytest.param(True, id="preallocated"),
    pytest.param(False, id="empty"),
])
@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_write_new(user_file, offset, size, preallocated):
    with io.open(user_file.path, "wb") as f:
        if preallocated:
            f.truncate(offset + size)

    src = io.BytesIO(b"x" * size)
    with file.open(user_file.url, "r+") as dst, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Write(dst, src, buf, size, offset=offset)
        op.run()

    with io.open(user_file.path, "rb") as f:
        # Nothing is written before offset.
        assert f.read(offset) == b"\0" * offset

        # All data was written.
        assert f.read(size) == src.getvalue()

        # Writing to unaligned size align file size by padding zeroes.
        file_size = os.path.getsize(user_file.path)
        trailer = file_size - offset - size
        assert file_size % user_file.sector_size == 0
        assert f.read() == b"\0" * trailer


@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_write_inside(user_file, offset, size):
    trailer = 8192
    with io.open(user_file.path, "wb") as f:
        f.truncate(offset + size + trailer)

    src = io.BytesIO(b"x" * size)
    with file.open(user_file.url, "r+") as dst, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Write(dst, src, buf, size, offset=offset)
        op.run()

    with io.open(user_file.path, "rb") as f:
        # Nothing is written before offset.
        assert f.read(offset) == b"\0" * offset

        # All data was written.
        assert f.read(size) == src.getvalue()

        # Nothing was written after offset + size, and file size is not
        # modified.
        assert f.read() == b"\0" * trailer


@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_write_partial_content(user_file, offset, size):
    with io.open(user_file.path, "wb") as f:
        f.truncate(size + offset)

    src = io.BytesIO(b"x" * (size - 1))
    with file.open(user_file.url, "r+") as dst, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Write(dst, src, buf, size, offset=offset)
        with pytest.raises(errors.PartialContent) as e:
            op.run()

    assert e.value.requested == size
    assert e.value.available == size - 1


def test_write_seek():
    dst = memory.Backend("r+", bytearray(b"a" * 10))
    dst.seek(8)
    src = io.BytesIO(b"b" * 5)
    with util.aligned_buffer(32) as buf:
        op = ops.Write(dst, src, buf, 5)
        op.run()
    dst.seek(0)
    b = bytearray(11)
    n = dst.readinto(b)
    assert n == 10
    assert b == b"bbbbbaaaaa\0"


@pytest.mark.parametrize("extra, dirty", [
    ({}, False),  # Flushes by default for backward compatibility.
    ({"flush": True}, False),
    ({"flush": False}, True),
])
def test_write_flush(extra, dirty):
    size = 4096
    dst = memory.Backend("r+", bytearray(b"a" * size))
    src = io.BytesIO(b"b" * size)
    with util.aligned_buffer(4096) as buf:
        op = ops.Write(dst, src, buf, size, **extra)
        op.run()
    assert dst.dirty == dirty


def test_write_repr():
    op = ops.Write(None, None, None, 100, offset=42)
    rep = repr(op)
    assert "Write" in rep
    assert "size=100 offset=42 done=0" in rep


def test_write_unbuffered_stream(user_file):
    chunks = [b"a" * 8192,
              b"b" * 42,
              b"c" * (8192 - 42)]
    src = util.UnbufferedStream(chunks)
    size = sum(len(c) for c in chunks)

    with file.open(user_file.url, "r+") as dst, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Write(dst, src, buf, size)
        op.run()

    with io.open(user_file.path, "rb") as f:
        for c in chunks:
            assert f.read(len(c)) == c
        assert f.read() == b""


def test_write_unbuffered_stream_partial_content(user_file):
    chunks = [b"a" * 8192,
              b"b" * 42,
              b"c" * (8192 - 42)]
    src = util.UnbufferedStream(chunks)
    size = sum(len(c) for c in chunks)

    with file.open(user_file.url, "r+") as dst, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Write(dst, src, buf, size + 1)
        with pytest.raises(errors.PartialContent):
            op.run()


@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_write_no_size(user_file, offset, size):
    with io.open(user_file.path, "wb") as f:
        f.truncate(offset + size)

    src = io.BytesIO(b"x" * size)
    with file.open(user_file.url, "r+") as dst, \
            util.aligned_buffer(1024**2) as buf:
        op = ops.Write(dst, src, buf, offset=offset)
        op.run()

    with io.open(user_file.path, "rb") as f:
        assert f.read(offset) == b"\0" * offset
        assert f.read(size) == src.getvalue()

        file_size = os.path.getsize(user_file.path)
        trailer = file_size - offset - size
        assert file_size % user_file.sector_size == 0
        assert f.read() == b"\0" * trailer


@pytest.mark.parametrize("sparse", [
    pytest.param(True, id="sparse"),
    pytest.param(False, id="preallocated"),
])
@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_zero_full(user_file, offset, size, sparse):
    data = b"x" * (offset + size)
    with io.open(user_file.path, "wb") as f:
        f.write(data)

    with file.open(user_file.url, "r+", sparse=sparse) as dst:
        op = ops.Zero(dst, size, offset=offset)
        op.run()

    with io.open(user_file.path, "rb") as f:
        # Nothing was zeroed before offset
        assert f.read(offset) == data[:offset]

        # Everything was zeroed after offset.
        assert f.read(size) == b"\0" * size

        # Zeroing to unaligned size align file size by padding zereos.
        file_size = os.path.getsize(user_file.path)
        assert file_size % user_file.sector_size == 0
        assert f.read() == b"\0" * (file_size - offset - size)


@pytest.mark.parametrize("sparse", [
    pytest.param(True, id="sparse"),
    pytest.param(False, id="preallocated"),
])
@pytest.mark.parametrize("offset,size", OFFSET_SIZE)
def test_zero_inside(user_file, offset, size, sparse):
    trailer = 8192
    data = b"x" * (offset + size + trailer)
    with io.open(user_file.path, "wb") as f:
        f.write(data)

    with file.open(user_file.url, "r+", sparse=sparse) as dst:
        op = ops.Zero(dst, size, offset=offset)
        op.run()

    with io.open(user_file.path, "rb") as f:
        # Nothing was zeroed before offset
        assert f.read(offset) == data[:offset]

        # Everything was zeroed after offset.
        assert f.read(size) == b"\0" * size

        # Nothing was zeroed after size.
        assert f.read() == data[offset + size:]


def test_zero_seek():
    dst = memory.Backend("r+", bytearray(b"a" * 10))
    dst.seek(8)
    op = ops.Zero(dst, 5)
    op.run()
    dst.seek(0)
    b = bytearray(11)
    n = dst.readinto(b)
    assert n == 10
    assert b == b"\0\0\0\0\0aaaaa\0"


@pytest.mark.parametrize("extra, dirty", [
    ({}, True),  # Does not flush by default.
    ({"flush": True}, False),
    ({"flush": False}, True),
])
def test_zero_flush(extra, dirty):
    size = 4096
    dst = memory.Backend("r+", bytearray(b"a" * size))
    op = ops.Zero(dst, size, **extra)
    op.run()
    assert dst.dirty == dirty


def test_zero_repr():
    op = ops.Zero(memory.Backend("r+"), 4096)
    rep = repr(op)
    assert "Zero" in rep
    assert "offset=0" in rep
    assert "size=4096" in rep
    assert "done=0" in rep


def test_flush():
    dst = memory.Backend("r+")
    dst.write(b"x")
    op = ops.Flush(dst)
    op.run()
    assert not dst.dirty


def test_flush_repr():
    op = ops.Flush(memory.Backend("r"))
    rep = repr(op)
    assert "Flush" in rep
    assert "done=0" in rep



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
cors - Cross-Origin Resource Sharing support

Browser send OPTIONS preflight request before sending other request to
check if the CORS protocol is supported, and the server allows the
method and headers.  The result of the preflight request may be cached
by the browser.

Here is an example preflight request sent during upload:

Host: example.com:54323
User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; ... Firefox/74.0
Accept: /
Accept-Language: en-US,en;q=0.5
Accept-Encoding: gzip, deflate, br
Access-Control-Request-Method: PUT
Access-Control-Request-Headers: cache-control,content-range,content-type
Referer: https://example.com:8443/ovirt-engine/webadmin/?locale=en_US
Origin: https://example.com:8443
Connection: keep-alive

The response must include these headers:

Access-Control-Allow-Origin: *
Access-Control-Allow-Headers: *
Access-Control-Allow-Methods: OPTIONS,GET,PUT
Access-Control-Max-Age: 86400

In this case we support any origin, any header and methods OPTIONS, GET,
and PUT.

For more info see:
https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS.
"""

import functools


def allow(allow_origin="*", allow_headers="*", allow_methods="*",
          max_age=24 * 3600):
    """
    Returns a decorator adding CORS headers to preflight request (OPTIONS) or
    actual request with a Origin header.

    Used for methods that may be called from engine webadmin.

    Example usage:

        @cors.allow()
        def get(self, req, resp):
            ...
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, req, resp, *args):
            modified = False

            # Allow only allow_origin.
            if req.headers.get("origin"):
                resp.headers["access-control-allow-origin"] = allow_origin
                modified = True

            # Allow only allow_headers.
            if req.headers.get("access-control-request-headers"):
                resp.headers["access-control-allow-headers"] = allow_headers
                modified = True

            # Allow only allow_methods.
            if req.headers.get("access-control-request-method"):
                resp.headers["access-control-allow-methods"] = allow_methods
                modified = True

            # Clients can cache the response for specified seconds.
            if modified:
                resp.headers["access-control-max-age"] = max_age

            return func(self, req, resp, *args)

        return wrapper

    return decorator



# ovirt-diskio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import io
import os
import shutil
import struct
import tarfile
import tempfile
import urllib.parse

import pytest

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import nbdutil
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd
from ovirt_imageio._internal import sockutil

from . import testutil
from . marks import flaky_in_ovirt_ci


@pytest.fixture
def tmpfs_dir():
    path = tempfile.mkdtemp(dir="/dev/shm")
    yield path
    shutil.rmtree(path)


@pytest.mark.parametrize("addr,export,url", [
    (nbd.UnixAddress("/sock"), None, "nbd:unix:/sock"),
    (nbd.UnixAddress("/sock"), "", "nbd:unix:/sock"),
    (nbd.UnixAddress("/sock"), "sda", "nbd:unix:/sock:exportname=sda"),
    (nbd.TCPAddress("host", 10900), None, "nbd:host:10900"),
    (nbd.TCPAddress("host", 10900), "", "nbd:host:10900"),
    (nbd.TCPAddress("host", 10900), "sdb", "nbd:host:10900:exportname=sdb"),
])
def test_server_url(addr, export, url):
    srv = qemu_nbd.Server("image", "raw", addr, export_name=export)
    assert srv.url == urllib.parse.urlparse(url)


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_open(tmpdir, fmt):
    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=1024**2)

    offset = 64 * 1024
    data = b"it works"

    with qemu_nbd.open(disk, fmt) as d:
        d.write(offset, data)
        d.flush()

    with qemu_nbd.open(disk, fmt, read_only=True) as d:
        assert d.read(offset, len(data)) == data


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_server_can_use_direct_io(tmpdir, nbd_server, fmt):
    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=1024**2)

    nbd_server.image = disk
    nbd_server.fmt = fmt
    nbd_server.cache = None
    nbd_server.aio = None

    # Starting probes the image and select cache="none" and aio="native"
    # if direct I/O can be used.
    nbd_server.start()
    assert nbd_server.cache == "none"
    assert nbd_server.aio == "native"


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_server_cannot_use_direct_io(tmpfs_dir, nbd_server, fmt):
    disk = os.path.join(tmpfs_dir, "disk." + fmt)
    qemu_img.create(disk, fmt, size=1024**2)

    nbd_server.image = disk
    nbd_server.fmt = fmt
    nbd_server.cache = None
    nbd_server.aio = None

    # Starting probes the image and selects cache="writeback" and
    # aio="threads" since direct I/O cannot be used with tmpfs file
    # system.
    nbd_server.start()
    assert nbd_server.cache == "writeback"
    assert nbd_server.aio == "threads"


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_server_keep_cache(tmpdir, nbd_server, fmt):
    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=1024**2)

    nbd_server.image = disk
    nbd_server.fmt = fmt
    nbd_server.cache = "writeback"
    nbd_server.aio = None

    # Staring the server selects aio="threads" since cache is not
    # "none".
    nbd_server.start()
    assert nbd_server.cache == "writeback"
    assert nbd_server.aio == "threads"


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_server_keep_aio(tmpdir, nbd_server, fmt):
    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=1024**2)

    nbd_server.image = disk
    nbd_server.fmt = fmt
    nbd_server.cache = None
    nbd_server.aio = "threads"

    # Staring the server probes the image and selects cache="none", but
    # keeps the requested aio option.
    nbd_server.start()
    assert nbd_server.cache == "none"
    assert nbd_server.aio == "threads"


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_server_keep_cache_and_aio(tmpdir, nbd_server, fmt):
    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=1024**2)

    nbd_server.image = disk
    nbd_server.fmt = fmt
    nbd_server.cache = "writeback"
    nbd_server.aio = "threads"

    # Staring the server keeps the requested options.
    nbd_server.start()
    assert nbd_server.cache == "writeback"
    assert nbd_server.aio == "threads"


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_ova(tmpdir, fmt):
    size = 1024**2
    offset = 64 * 1024
    disks = []

    # Created disks with unique content.
    for i in range(2):
        disk = str(tmpdir.join("disk{}.{}".format(i, fmt)))
        qemu_img.create(disk, fmt, size=size)
        with qemu_nbd.open(disk, fmt) as d:
            d.write(offset, disk.encode("utf-8"))
            d.flush()
        disks.append(disk)

    ova = str(tmpdir.join("vm.ova"))

    # Create a ova file.
    with tarfile.open(ova, "w") as tar:
        for disk in disks:
            tar.add(disk, arcname=os.path.basename(disk))

    # Read disks contents from the ova file.
    with tarfile.open(ova) as tar:
        for disk in disks:
            member = tar.getmember(os.path.basename(disk))
            with qemu_nbd.open(
                    ova,
                    fmt=fmt,
                    read_only=True,
                    offset=member.offset_data,
                    size=member.size) as d:
                assert d.export_size == size
                data = disk.encode("utf-8")
                assert d.read(offset, len(data)) == data


def test_ova_compressed_qcow2(tmpdir):
    size = 1024**2
    offset = 64 * 1024
    data = b"I can eat glass and it doesn't hurt me."

    tmp = str(tmpdir.join("disk.raw"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.seek(offset)
        f.write(data)

    disk = str(tmpdir.join("disk.qcow2"))
    qemu_img.convert(tmp, disk, "raw", "qcow2", compressed=True)

    ova = str(tmpdir.join("vm.ova"))

    # Create tar file with compressed qcow2 disk.
    with tarfile.open(ova, "w") as tar:
        tar.add(disk, arcname=os.path.basename(disk))

    # Read disk contents from the tar file.
    with tarfile.open(ova) as tar:
        member = tar.getmember(os.path.basename(disk))
        with qemu_nbd.open(
                ova,
                fmt="qcow2",
                read_only=True,
                offset=member.offset_data,
                size=member.size) as d:
            assert d.export_size == size
            assert d.read(offset, len(data)) == data


def test_run_unix(tmpdir):
    image = str(tmpdir.join("image"))
    sock = str(tmpdir.join("sock"))

    with io.open(image, "wb") as f:
        f.truncate(1024**2)

    addr = nbd.UnixAddress(sock)

    with qemu_nbd.run(image, "raw", addr):
        # The helper already waited for the NBD socket, not wait needed.
        assert sockutil.wait_for_socket(addr, 0.0)

    # The socket must be closed, no wait needed.
    assert not sockutil.wait_for_socket(addr, 0.0)


@flaky_in_ovirt_ci
def test_run_tcp(tmpfile):
    with io.open(tmpfile, "r+b") as f:
        f.truncate(1024**2)

    addr = nbd.TCPAddress("localhost", testutil.random_tcp_port())

    with qemu_nbd.run(tmpfile, "raw", addr):
        # The helper already waited for the NBD socket, not wait needed.
        assert sockutil.wait_for_socket(addr, 0.0)

    # The socket must be closed, no wait needed.
    assert not sockutil.wait_for_socket(addr, 0.0)


@pytest.mark.parametrize("options", [
    pytest.param({
        "cache": "none",
        "aio": "native",
        "discard": "unmap",
        "detect_zeroes": "unmap"
    }, id="qemu_nbd defaults"),
    pytest.param({
        "cache": None,
        "aio": None,
        "discard": None,
        "detect_zeroes": None
    }, id="qemu-nbd defaults"),
])
@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_options(tmpdir, fmt, options):
    size = 4 * 1024**2
    chunk_size = 128 * 1024

    src = str(tmpdir.join("src." + fmt))
    qemu_img.create(src, fmt, size=size)

    with qemu_nbd.open(src, fmt) as c:
        for offset in range(0, size, chunk_size):
            c.write(offset, struct.pack(">Q", offset))
        c.flush()

    dst = str(tmpdir.join("dst." + fmt))
    qemu_img.create(dst, fmt, size=size)

    src_addr = nbd.UnixAddress(str(tmpdir.join("src.sock")))
    dst_addr = nbd.UnixAddress(str(tmpdir.join("dst.sock")))

    with qemu_nbd.run(
                src, fmt, src_addr,
                read_only=True,
                **options), \
            qemu_nbd.run(
                dst, fmt, dst_addr,
                **options), \
            nbd.Client(src_addr) as src_client, \
            nbd.Client(dst_addr) as dst_client:

        nbdutil.copy(src_client, dst_client)

    qemu_img.compare(src, dst)


def test_backing_chain(tmpdir):
    size = 128 * 1024
    base = str(tmpdir.join("base.raw"))
    top = str(tmpdir.join("top.qcow2"))

    base_data = b"data from base".ljust(32, b"\0")

    # Add base image with some data.
    qemu_img.create(base, "raw", size=size)
    with qemu_nbd.open(base, "raw") as c:
        c.write(0, base_data)
        c.flush()

    # Add empty overlay.
    qemu_img.create(top, "qcow2", backing_file=base, backing_format="raw")

    top_addr = nbd.UnixAddress(str(tmpdir.join("sock")))

    # By default, we see data from base.
    with qemu_nbd.run(top, "qcow2", top_addr), \
            nbd.Client(top_addr) as c:
        assert c.read(0, 32) == base_data

    # With backing chain disabled, we see data only from top.
    with qemu_nbd.run(top, "qcow2", top_addr, backing_chain=False), \
            nbd.Client(top_addr) as c:
        assert c.read(0, 32) == b"\0" * 32


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_shared(tmpdir, fmt):
    size = 1024**2
    chunk_size = size // 2

    src = str(tmpdir.join("src." + fmt))
    qemu_img.create(src, fmt, size=size)

    with qemu_nbd.open(src, fmt) as c:
        c.write(0, b"a" * chunk_size)
        c.write(0, b"b" * chunk_size)
        c.flush()

    dst = str(tmpdir.join("dst." + fmt))
    qemu_img.create(dst, fmt, size=size)

    src_addr = nbd.UnixAddress(str(tmpdir.join("src.sock")))
    dst_addr = nbd.UnixAddress(str(tmpdir.join("dst.sock")))

    # Start 2 qemu-nbd servers, each with 2 clients that can read and write in
    # parallel for higher throughput.

    with qemu_nbd.run(src, fmt, src_addr, read_only=True, shared=2), \
            qemu_nbd.run(dst, fmt, dst_addr, shared=2), \
            nbd.Client(src_addr) as src_client1, \
            nbd.Client(src_addr) as src_client2, \
            nbd.Client(dst_addr) as dst_client1, \
            nbd.Client(dst_addr) as dst_client2:

        # Copy first half of the image with src_client1 and dst_client2 and
        # second half with src_client2 and dst_client2. In a real application
        # we would have more clients, running in multiple threads.

        chunk1 = src_client1.read(0, chunk_size)
        dst_client1.write(0, chunk1)

        chunk2 = src_client2.read(chunk_size, chunk_size)
        dst_client2.write(chunk_size, chunk2)

        dst_client1.flush()
        dst_client2.flush()

    qemu_img.compare(src, dst)


def test_dirty_bitmap(tmpdir):
    size = 1024**2

    # Create image with empty bitmap.
    img = str(tmpdir.join("img.qcow2"))
    qemu_img.create(img, "qcow2", size=size)
    qemu_img.bitmap_add(img, "b0")

    # Write data to image, modifying the bitmap.
    with qemu_nbd.open(img, "qcow2") as c:
        # This will allocate one cluster. By default bitmap granularity is also
        # one cluster, so this will make the first extent dirty.
        c.write(0, b"a")
        c.flush()

    # Read dirty extents.
    with qemu_nbd.open(img, "qcow2", read_only=True, bitmap="b0") as c:
        extents = c.extents(0, size)[nbd.QEMU_DIRTY_BITMAP + "b0"]

    bitmap = qemu_img.info(img)["format-specific"]["data"]["bitmaps"][0]

    assert extents == [
        nbd.Extent(length=bitmap["granularity"], flags=nbd.EXTENT_DIRTY),
        nbd.Extent(length=size - bitmap["granularity"], flags=0),
    ]


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
@pytest.mark.parametrize("detect_zeroes", ["on", "unmap"])
def test_detect_zeroes_enabled(tmpdir, fmt, detect_zeroes):
    size = 1024**2

    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=size)

    with qemu_nbd.open(disk, fmt, detect_zeroes=detect_zeroes) as c:
        # These zeroes should be detected.
        c.write(0, b"\0" * size)
        c.flush()
        extents = c.extents(0, size)

    assert extents["base:allocation"] == [
        nbd.Extent(length=1048576, flags=3),
    ]

    if fmt != "raw":
        assert extents["qemu:allocation-depth"] == [
            nbd.Extent(length=1048576, flags=0),
        ]


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
@pytest.mark.parametrize("detect_zeroes", ["off", None])
def test_detect_zeroes_disabled(tmpdir, fmt, detect_zeroes):
    size = 1024**2

    disk = str(tmpdir.join("disk." + fmt))
    qemu_img.create(disk, fmt, size=size)

    with qemu_nbd.open(disk, fmt, detect_zeroes=detect_zeroes) as c:
        # These zeroes should not be detected.
        c.write(0, b"\0" * size)
        c.flush()
        extents = c.extents(0, size)

    assert extents["base:allocation"] == [
        nbd.Extent(length=1048576, flags=0),
    ]

    if fmt != "raw":
        assert extents["qemu:allocation-depth"] == [
            nbd.Extent(length=1048576, flags=0),
        ]



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import functools
import logging
import os
import re
import subprocess

from contextlib import contextmanager

from ovirt_imageio._internal import sockutil

QEMU = os.environ.get("QEMU", "qemu-kvm")

log = logging.getLogger("qemu")


def supports_audiodev():
    if not hasattr(supports_audiodev, "result"):
        cmd = [QEMU, "--help"]
        out = subprocess.check_output(cmd, env=env()).decode()
        m = re.search(r"^-audiodev +none\b", out, flags=re.MULTILINE)
        supports_audiodev.result = m is not None
    return supports_audiodev.result


def env():
    """
    Amend PATH to locate qemu-kvm on platforms that hide it in /usr/libexec
    (e.g RHEL).
    """
    env = dict(os.environ)
    env["PATH"] = ":".join((env["PATH"], "/usr/libexec"))
    return env


@contextmanager
def run(image, fmt, qmp_sock=None, start_cpu=True, shutdown_timeout=1):
    # NOTES:
    # - Let qemu pick default memory size, since on some platforms memory have
    #   strange alignment. Here is a failure from ppc64le host:
    #       qemu-kvm: Memory size 0x1000000 is not aligned to 256 MiB
    cmd = [
        QEMU,
        # Use kvm if available, othrewise fallback to tcg. This allows running
        # qemu on Travis CI which does not support nested virtualization.
        "-nodefaults",
        "-machine", "accel=kvm:tcg",
        "-drive",
        "if=virtio,id=drive0,node-name=file0,file={},format={}".format(
            image, fmt),
        "-nographic",
        "-net", "none",
        "-monitor", "none",
        "-serial", "stdio",
    ]

    if qmp_sock:
        cmd.append("-qmp")
        cmd.append("unix:{},server,nowait".format(qmp_sock))

    # Workaround for bug in qemu-4.0.0-rc0 on Fedora, failing to start VM
    # becuase initilizing real audio driver failed.
    # See https://bugzilla.redhat.com/1692047.
    if supports_audiodev():
        cmd.append("-audiodev")
        cmd.append("none,id=1")

    if not start_cpu:
        cmd.append("-S")

    log.info("Starting qemu %s", cmd)
    p = subprocess.Popen(
        cmd,
        env=env(),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE)
    try:
        if qmp_sock and not sockutil.wait_for_socket(qmp_sock, 1):
            raise RuntimeError("Timeout waiting for socket: %s" % qmp_sock)
        yield Guest(p)
    finally:
        log.info("Terminating qemu gracefully")
        p.terminate()
        try:
            p.wait(shutdown_timeout)
        except subprocess.TimeoutExpired:
            log.warning("Timeout terminating qemu - killing it")
            p.kill()
            p.wait()
        log.debug("qemu terminated with exit code %s", p.returncode)


class Guest:

    def __init__(self, p):
        self._stdin = p.stdin
        self._stdout = p.stdout
        self._logged = False

    def login(self, name, password):
        log.info("Logging in to guest")
        assert not self._logged
        self._wait_for("login: ")
        self._send(name)
        self._wait_for("Password: ")
        self._send(password)
        self._wait_for("# ")
        self._logged = True

    def run(self, command):
        log.info("Running command in guest: %s", command)
        self._send(command)
        return self._wait_for("# ")

    def _send(self, s):
        log.debug("Sending: %r", s)
        self._stdin.write(s.encode("utf-8") + b"\n")
        self._stdin.flush()
        self._wait_for(s + "\r\n")

    def _wait_for(self, s):
        log.debug("Waiting for: %r", s)
        data = s.encode("utf-8")
        buf = bytearray()
        while True:
            buf += self._stdout.read(1)
            if buf.endswith(data):
                rep = buf[:-len(data)]
                return rep.decode("utf-8")


@functools.lru_cache(maxsize=1)
def version():
    # Typical output:
    # QEMU emulator version 6.0.0 (qemu-6.0.0-1.fc32)
    # Copyright (c) 2003-2021 Fabrice Bellard and the QEMU Project developers
    out = subprocess.check_output([QEMU, "--version"], env=env())
    first_line = out.decode("utf-8").splitlines()[0]
    _, _, _, version_string, pkg = first_line.split(None, 4)
    return tuple(int(n) for n in version_string.split("."))



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import hashlib
import logging

from .. import backends
from .. import blkhash
from .. import errors
from .. import http
from .. import ioutil
from .. import ops
from .. import util
from .. import validate

log = logging.getLogger("checksum")

ALGORITHMS = frozenset(hashlib.algorithms_available)

# Limit allowed block size to avoid abusing server resources.
MIN_BLOCK_SIZE = blkhash.BLOCK_SIZE // 4
MAX_BLOCK_SIZE = blkhash.BLOCK_SIZE * 4


class Checksum:
    """
    Handle requests for the /images/ticket-id/checksum resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def get(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        algorithm = validate.enum(
            req.query,
            "algorithm",
            ALGORITHMS,
            default=blkhash.ALGORITHM)

        try:
            block_size = int(req.query.get("block_size", blkhash.BLOCK_SIZE))
        except ValueError:
            raise http.Error(
                http.BAD_REQUEST,
                "Invalid block size: {!r}".format(req.query["block_size"]))

        if not MIN_BLOCK_SIZE <= block_size <= MAX_BLOCK_SIZE:
            raise http.Error(
                http.BAD_REQUEST,
                "Block size out of allowed range: {}-{}"
                .format(MIN_BLOCK_SIZE, MAX_BLOCK_SIZE))

        if block_size % 4096:
            raise http.Error(
                http.BAD_REQUEST, "Block size is not aligned to 4096")

        try:
            ticket = self.auth.authorize(ticket_id, "read")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            raise http.Error(http.FORBIDDEN, str(e))

        log.info("[%s] CHECKSUM transfer=%s algorithm=%s block_size=%s",
                 req.client_addr, ticket.transfer_id, algorithm, block_size)

        # For simplicity we create a new buffer even if block_size is same as
        # ctx.buffer length.

        with util.aligned_buffer(block_size) as buf:
            op = Operation(ctx.backend, buf, algorithm, clock=req.clock)
            try:
                checksum = ticket.run(op)
            except errors.AuthorizationError as e:
                resp.close_connection()
                raise http.Error(http.FORBIDDEN, str(e)) from None

        resp.send_json(checksum)


class Algorithms:
    """
    Handle requests for the /images/ticket-id/checksum/algorithms resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def get(self, req, resp, *args):
        # Server information, no authorization needed.
        resp.send_json({"algorithms": sorted(ALGORITHMS)})


class Operation(ops.Operation):
    """
    Checksum operation.
    """

    name = "checksum"

    def __init__(self, backend, buf, algorithm, detect_zeroes=True,
                 clock=None):
        super().__init__(size=backend.size(), buf=buf, clock=clock)
        self._backend = backend
        self._algorithm = algorithm
        self._detect_zeroes = detect_zeroes

    def _run(self):
        block_size = len(self._buf)
        # Only blakse2b and blake2s support variable digest size, and 32 works
        # with both and is large enough.
        if self._algorithm.startswith("blake2"):
            digest_size = blkhash.DIGEST_SIZE
        else:
            digest_size = None

        h = blkhash.Hash(
            block_size=block_size,
            algorithm=self._algorithm,
            digest_size=digest_size)

        for block in blkhash.split(self._backend.extents("zero"), block_size):
            if block.zero:
                h.zero(block.length)
            else:
                with memoryview(self._buf)[:block.length] as view:
                    self._backend.seek(block.start)
                    self._backend.readinto(view)
                    if self._detect_zeroes and ioutil.is_zero(view):
                        h.zero(block.length)
                    else:
                        h.update(view)

            self._done += block.length

            if self._canceled:
                raise ops.Canceled

        return {
            "algorithm": self._algorithm,
            "block_size": block_size,
            "checksum": h.hexdigest(),
        }


def compute(backend, buf, algorithm=blkhash.ALGORITHM, detect_zeroes=True):
    """
    Compute image checksum.
    """
    op = Operation(backend, buf, algorithm, detect_zeroes=detect_zeroes)
    return op.run()



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import threading
import time

from .. _internal import util


class ProgressBar:

    def __init__(self, size=0, output=sys.stdout, step=0.1, width=79,
                 now=time.monotonic):
        """
        Argumnets:
            size (int): total number of bytes. If size is unknown when
                creating, progress value is not displayed. The size can be set
                later to enable progress display.
            output (fileobj): file to write progress to (default sys.stdout).
            step (float): mininum progress update interval in seconds (default
                0.1 seconds).
            width (int): width of progress bar in characters (default 79)
            now (callable): callable returning current time for testing.
        """
        self.size = size
        self.output = output
        self.step = step
        # TODO: use current terminal width instead.
        self.width = width
        self.now = now
        self.lock = threading.Lock()
        self.start = self.now()
        self.next = 0
        self.done = 0

        # The first update can take some time.
        self.update(0)

    def update(self, n):
        """
        Increment the progress by n bytes.

        Should be called for every successful transfer, with the number of
        bytes transfer. The number of bytes transfered is updated on every
        call, but the progress is updated only if step seconds have passed
        since the last update.

        Note: this interface is compatible with tqdm[1], to allow user to use
        different progress implementations.

        [1] https://github.com/tqdm/tqdm#manual
        """
        with self.lock:
            self.done += n
            now = self.now()
            if now < self.next:
                return

            self.next = now + self.step
            self._draw(now)

    def close(self):
        with self.lock:
            # If we wrote progress, we need to draw the last progress line.
            if self.done > 0:
                self._draw(self.now(), last=True)

    def _draw(self, now, last=False):
        elapsed = now - self.start

        if self.size:
            progress = "%6.2f%%" % (self.done / self.size * 100)
        else:
            progress = "-------"

        line = "[ %s ] %s, %.2f seconds, %s/s" % (
            progress,
            util.humansize(self.done),
            elapsed,
            util.humansize(self.done / elapsed if elapsed else 0),
        )

        line = line.ljust(self.width, " ")

        # Using "\r" moves the cursor to the first column, so the next progress
        # will overwite this one. If this is the last progress, we use "\n" to
        # move to the next line. Otherwise, the next shell prompt will include
        # part of the old progress.
        end = "\n" if last else "\r"

        self.output.write(line + end)
        self.output.flush()

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest

from ovirt_imageio._internal import configloader


@pytest.fixture
def config():
    class config:
        class foo:
            string = "old"
            string_nonascii = "\u05d0"
            string_null = "\u0000"
            integer = 1
            real = 4.0
            boolean = False

        class bar:
            string = "old"
            keyword__class = "bar.class"
    return config


def test_empty(tmpdir, config):
    conf = str(tmpdir.join("conf"))
    with open(conf, "w"):
        pass
    configloader.load(config, [conf])
    assert config.foo.string == "old"
    assert config.foo.string_nonascii == "\u05d0"
    assert config.foo.string_null == "\u0000"
    assert config.foo.integer == 1
    assert config.foo.real == 4.0
    assert config.foo.boolean is False
    assert config.bar.string == "old"


def test_ignore_unknown_section(tmpdir, config):
    data = """
[foo]
string = new

[unknown]
string = new
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.foo.string == "new"
    assert config.foo.integer == 1
    assert config.foo.real == 4.0
    assert config.foo.boolean is False
    assert config.bar.string == "old"
    assert not hasattr(config, "unknown")


def test_ignore_unknown_option(tmpdir, config):
    data = """
[foo]
string = new
unknown = 3
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.foo.string == "new"
    assert config.foo.integer == 1
    assert config.foo.real == 4.0
    assert config.foo.boolean is False
    assert config.bar.string == "old"
    assert not hasattr(config.foo, "unknown")


def test_some(tmpdir, config):
    data = """
[foo]
string = new
boolean = true
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.foo.string == "new"
    assert config.foo.integer == 1
    assert config.foo.real == 4.0
    assert config.foo.boolean is True
    assert config.bar.string == "old"


def test_full(tmpdir, config):
    data = """
[foo]
string = new
integer = 2
real = 4.1
boolean = true

[bar]
string = new
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.foo.string == "new"
    assert config.foo.integer == 2
    assert config.foo.real == 4.1
    assert config.foo.boolean is True
    assert config.bar.string == "new"


@pytest.mark.parametrize("value", [
    "True", "tRue", "true",
    "Yes", "yeS", "yes",
    "On", "oN", "on",
    "1",
])
def test_true(tmpdir, config, value):
    data = """
[foo]
boolean = %s
""" % value
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.foo.boolean is True


@pytest.mark.parametrize("value", [
    "False", "faLse", "false",
    "No", "nO", "no",
    "Off", "ofF", "off",
    "0",
])
def test_false(tmpdir, config, value):
    data = """
[foo]
boolean = %s
""" % value
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.foo.boolean is False


@pytest.mark.parametrize("option", ["integer", "real", "boolean"])
def test_validate(tmpdir, config, option):
    data = """
[foo]
%s = invalid value
""" % option
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    pytest.raises(ValueError, configloader.load, config, [conf])


def test_unicode(tmpdir, config):
    data = """
[foo]
string_nonascii = \u05d0
string_null = \u0000
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "wb") as f:
        f.write(data.encode("utf-8"))
    configloader.load(config, [conf])
    assert config.foo.string_nonascii == "\u05d0"
    assert config.foo.string_null == "\u0000"


def test_unsupported_default_value(tmpdir):

    class config:
        class section:
            value = b"bytes"

    data = """
[section]
value = bar
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)

    with pytest.raises(ValueError) as e:
        configloader.load(config, [conf])

    error = str(e.value)
    assert "section.value" in error
    assert str(type(config.section.value)) in error


def test_keyword_option(tmpdir, config):
    data = """
[bar]
string = new
class = test.class
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    assert config.bar.string == "new"
    assert config.bar.keyword__class == "test.class"


def test_config_to_dict(tmpdir, config):
    data = """
[foo]
string = new
integer = 2
real = 4.1
boolean = true

[bar]
string = new
class = test.class
"""
    conf = str(tmpdir.join("conf"))
    with open(conf, "w") as f:
        f.write(data)
    configloader.load(config, [conf])
    expected = {
        "foo": {
            "string": "new",
            "integer": 2,
            "real": 4.1,
            "boolean": True,
            "string_nonascii": "\u05d0",
            "string_null": "\u0000",
        },
        "bar": {
            "string": "new",
            "class": "test.class"
        },
    }
    assert expected == configloader.to_dict(config)



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
configloader - simpler configuration loader

This module loads configuration files using ini file format, validates
options and updates given config module.

To load configuration, define the configuration structure in a module,
for example config.py:

    # config.py

    class foo:

        string = "value"
        integer = 42
        real = 3.14
        boolean = False

    class bar:

        boolean = True

The configuration file should match the class structure:

    # /etc/app/app.conf

    [foo]
    string = new
    integer = 43
    real = 3.15
    boolean = true

    [bar]
    boolean = false

To load the configuration from "/etc/app/app.conf", use:

    import config
    import configloader
    ...
    configloader.load(config, ["/etc/app/app.conf"])
    assert config.foo.integer == 43

Invalid values in the configuration file that do not match the types in
the config module will raise ValueError.

Unknown sections and options in the configuration file are ignored.

String values must use unicode literals (e.g. "value"). bytes values (.e.g
b"value" or "value") are not supported.

In case that using python keyword as an option is required, corresponding class
attribute has to have prefix "keyword__", e.g. to load config

    # /etc/app/app.conf

    [foo]
    class = my.app.class

appropriate config class should be

    # config.py

    class foo:

        keyword__class = "default.class"
"""

import configparser
import keyword

KEYWORD_PREFIX = "keyword__"


def keyword_mapping(option):
    option = option.lower()
    if keyword.iskeyword(option):
        option = "{}{}".format(KEYWORD_PREFIX, option)
    return option


def load(config, files):
    parser = configparser.RawConfigParser()
    parser.optionxform = keyword_mapping
    parser.read(files)
    for section_name in _public_names(config):
        section = getattr(config, section_name)
        for option in _public_names(section):
            try:
                value = parser.get(section_name, option)
            except configparser.NoSectionError:
                break
            except configparser.NoOptionError:
                continue

            value_type = type(getattr(section, option))
            if value_type not in _validators:
                raise ValueError(
                    "Unsupported default value type for '{}.{}': {}"
                    .format(section_name, option, value_type))

            validate = _validators[value_type]
            value = validate(value)
            setattr(section, option, value)


def to_dict(config):
    cfg_dict = {}
    for section_name in _public_names(config):
        section = getattr(config, section_name)
        cfg_dict[section_name] = _obj_to_dict(section)
    return cfg_dict


def _public_names(obj):
    return [name for name in dir(obj) if not name.startswith("_")]


def _obj_to_dict(obj):
    obj_dict = {}
    for attr in _public_names(obj):
        value = getattr(obj, attr)
        if attr.startswith(KEYWORD_PREFIX):
            attr = attr[len(KEYWORD_PREFIX):]
        obj_dict[attr] = value
    return obj_dict


def _validate_bool(s):
    # Use the same values configparser accepts
    val = s.lower()
    if val in ("true", "yes", "1", "on"):
        return True
    elif val in ("false", "no", "0", "off"):
        return False
    raise ValueError("Invalid boolean value: %r" % s)


_validators = {
    str: str,
    int: int,
    float: float,
    bool: _validate_bool,
}



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Smart client for accessing imageio server.

This module exposes the public names. Anything else in this package is private
and should not be used.
"""

# flake8: noqa

# Use for default buffer size.
from .. _internal.io import BUFFER_SIZE

from .. _internal import version

# The public APIs
from . _api import (
    upload,
    download,
    info,
    measure,
    checksum,
    extents,
    ImageioClient,
)

# For better user experience.
from . _ui import ProgressBar

__all__ = (
    "BUFFER_SIZE",
    "ImageioClient",
    "ProgressBar",
    "checksum",
    "download",
    "extents",
    "info",
    "measure",
    "upload",
)

__version__ = version.string



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
ipv6 - IP v6 helper module
"""


def unquote_address(host):
    """
    Unquote IPv6 quoted numeric address.
    """
    if host.startswith("[") and host.endswith("]") and ":" in host:
        host = host[1:-1]
    return host


def quote_address(host):
    """
    Quote IPv6 numeric address.
    """
    if ":" in host:
        host = "[{}]".format(host)
    return host



# ovirt-diskio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import io
import logging

import pytest

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd

from . import backup
from . import ci
from . import distro
from . import qemu
from . import qmp
from . import testutil

from . marks import flaky_in_ovirt_ci

log = logging.getLogger("test")


@pytest.mark.parametrize("transport", [
    "unix",
    pytest.param("tcp", marks=flaky_in_ovirt_ci),
])
@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_full_backup(tmpdir, fmt, transport):
    disk_size = 1024**2
    disk_part = disk_size // 4
    disk = str(tmpdir.join("disk." + fmt))
    backup_disk = str(tmpdir.join("backup.raw"))

    # Create disk
    qemu_img.create(disk, fmt, size=disk_size)

    # Pupulate disk with data.
    with qemu_nbd.open(disk, fmt) as d:
        for i in range(0, disk_size, disk_part):
            data = b"%d\n" % i
            d.write(i, data.ljust(512))
        d.flush()

    if transport == "unix":
        nbd_sock = nbd.UnixAddress(tmpdir.join("nbd.sock"))
    else:
        nbd_sock = nbd.TCPAddress("localhost", testutil.random_tcp_port())

    # Backup using qemu-img convert.
    with backup.full_backup(tmpdir, disk, fmt, nbd_sock):
        log.debug("Backing up image with qemu-img")
        qemu_img.convert(
            nbd_sock.url("sda"),
            backup_disk,
            src_fmt="raw",
            dst_fmt="raw",
            progress=True)

    # Compare source and backup disks.
    with qemu_nbd.open(disk, fmt, read_only=True) as d, \
            io.open(backup_disk, "rb") as b:
        for i in range(0, disk_size, disk_part):
            b.seek(i)
            assert d.read(i, 512) == b.read(512)


# This can take more than 30 seoconds when running on Travis without hardware
# acceleration:
# 41.77s setup    test/backup_test.py::test_full_backup_guest
# 17.59s call     test/backup_test.py::test_full_backup_guest
@pytest.mark.timeout(120)
@pytest.mark.xfail(
    ci.is_ovirt() or distro.is_centos("9"),
    reason="Always times out", run=False)
def test_full_backup_guest(tmpdir, base_image):
    base = qemu_img.info(base_image)
    disk_size = base["virtual-size"]

    disk = str(tmpdir.join("disk.qcow2"))
    qemu_img.create(
        disk, "qcow2", backing_file=base_image, backing_format=base["format"])

    scratch_disk = str(tmpdir.join("scratch.qcow2"))
    qemu_img.create(scratch_disk, "qcow2", size=disk_size)

    backup_disk = str(tmpdir.join("backup.qcow2"))
    qemu_img.create(backup_disk, "qcow2", size=disk_size)

    qmp_sock = nbd.UnixAddress(tmpdir.join("qmp.sock"))
    nbd_sock = nbd.UnixAddress(tmpdir.join("nbd.sock"))

    with qemu.run(disk, "qcow2", qmp_sock, shutdown_timeout=10) as guest, \
            qmp.Client(qmp_sock) as qmp_client:
        guest.login("root", "")

        assert guest.run("touch before-backup; sync") == ""

        with backup.run(
                qmp_client, nbd_sock, scratch_disk, checkpoint="check1"):

            assert guest.run("touch during-backup; sync") == ""

            backup.copy_disk(nbd_sock.url("sda"), backup_disk)

    verify_backup(backup_disk, ["before-backup"])


@pytest.mark.timeout(120)
@pytest.mark.xfail(
    ci.is_ovirt() or distro.is_centos("9"),
    reason="Always times out", run=False)
def test_incremental_backup_guest(tmpdir, base_image):
    base = qemu_img.info(base_image)
    disk_size = base["virtual-size"]

    disk = str(tmpdir.join("disk.qcow2"))
    qemu_img.create(
        disk, "qcow2", backing_file=base_image, backing_format=base["format"])

    scratch_disk = str(tmpdir.join("scratch.qcow2"))
    qemu_img.create(scratch_disk, "qcow2", size=disk_size)

    full_backup_disk = str(tmpdir.join("full-backup.qcow2"))
    qemu_img.create(full_backup_disk, "qcow2", size=disk_size)

    incr_backup_disk = str(tmpdir.join("incr-backup.qcow2"))
    qemu_img.create(incr_backup_disk, "qcow2", size=disk_size)

    qmp_sock = nbd.UnixAddress(tmpdir.join("qmp.sock"))
    nbd_sock = nbd.UnixAddress(tmpdir.join("nbd.sock"))

    with qemu.run(disk, "qcow2", qmp_sock, shutdown_timeout=10) as guest, \
            qmp.Client(qmp_sock) as qmp_client:
        guest.login("root", "")

        with backup.run(
                qmp_client, nbd_sock, scratch_disk, checkpoint="check1"):

            backup.copy_disk(nbd_sock.url("sda"), full_backup_disk)

        qemu_img.create(scratch_disk, "qcow2", size=disk_size)

        assert guest.run("touch before-backup; sync") == ""

        with backup.run(
                qmp_client, nbd_sock, scratch_disk, checkpoint="check2",
                incremental="check1"):

            assert guest.run("touch during-backup; sync") == ""

            backup.copy_dirty(nbd_sock.url("sda"), incr_backup_disk)

    qemu_img.unsafe_rebase(incr_backup_disk, full_backup_disk, "qcow2")
    verify_backup(incr_backup_disk, ["before-backup"])


def verify_backup(backup_disk, expected_files):
    log.info("Verifying backup")

    preview_disk = backup_disk + ".preview"
    qemu_img.create(
        preview_disk,
        "qcow2",
        backing_file=backup_disk,
        backing_format="qcow2")

    with qemu.run(preview_disk, "qcow2") as guest:
        guest.login("root", "")
        out = guest.run("ls -1 --color=never")
        assert out.splitlines() == expected_files



# ovirt-diskio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import logging
import time

from contextlib import contextmanager

from urllib.parse import urlparse

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import nbdutil
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd

from . import qemu
from . import qmp

log = logging.getLogger("backup")


@contextmanager
def full_backup(tmpdir, disk, fmt, sock, checkpoint=None):
    """
    Start qemu internal nbd server using address sock, exposing disk for
    full backup, creating temporary files in tmpdir.
    """
    scratch_disk = str(tmpdir.join("scratch.qcow2"))
    qmp_sock = nbd.UnixAddress(tmpdir.join("qmp.sock"))

    disk_size = qemu_img.info(disk)["virtual-size"]
    qemu_img.create(scratch_disk, "qcow2", size=disk_size)

    with qemu.run(disk, fmt, qmp_sock, start_cpu=False, shutdown_timeout=10), \
            qmp.Client(qmp_sock) as c, \
            run(c, sock, scratch_disk, checkpoint=checkpoint):
        yield


@contextmanager
def run(qmp, sock, scratch_disk, checkpoint=None, incremental=None):
    b = Backup(qmp, sock, scratch_disk, checkpoint=checkpoint,
               incremental=incremental)
    b.start()
    try:
        yield b
    finally:
        b.stop()


class Backup:

    def __init__(self, qmp, sock, scratch_disk, checkpoint=None,
                 incremental=None):
        self.qmp = qmp
        self.sock = sock
        self.scratch_disk = scratch_disk
        self.checkpoint = checkpoint
        self.incremental = incremental

        # Hardcoded value, good enough for now.
        self.export = "sda"

        # Use node name "file0" as a stable reference to our disk. It will not
        # change when the block graph is modifed during backup.
        self.file = "file0"

        self.job = "job0"

        # Libvirt uses something like "backup-libvirt-42-format".
        self.node = "backup-file0"
        self.bitmap = self.node if self.incremental else None

    def start(self):
        log.info("Starting backup checkpoint=%s incremental=%s",
                 self.checkpoint, self.incremental)

        self.start_nbd_server()
        self.add_backup_node()

        if self.incremental:
            self.add_dirty_bitmap()

        self.run_backup_transaction()
        self.add_to_nbd_server()

    def stop(self):
        log.info("Stopping backup checkpoint=%s incremental=%s",
                 self.checkpoint, self.incremental)

        # Give qemu time to detect that the NBD client disconnected before
        # we tear down the nbd server.
        log.debug("Waiting before tearing down nbd server")
        time.sleep(0.1)

        self.remove_from_nbd_server()
        self.stop_nbd_server()
        self.cancel_block_job()
        self.remove_backup_node()

        if self.incremental:
            self.remove_dirty_bitmap()

    def start_nbd_server(self):
        log.debug("Starting nbd server listening on %s", self.sock)

        if self.sock.transport == "unix":
            addr = {"type": "unix",
                    "data": {"path": self.sock.path}}
        elif self.sock.transport == "tcp":
            addr = {"type": "inet",
                    "data": {"host": self.sock.host,
                             "port": str(self.sock.port)}}
        else:
            raise RuntimeError("Unsupported transport: {}".format(self.sock))

        self.qmp.execute("nbd-server-start", {"addr": addr})

    def add_backup_node(self):
        log.debug("Adding backup node %s for %s", self.node, self.file)

        self.qmp.execute("blockdev-add", {
            "driver": "qcow2",
            "node-name": self.node,
            "file": {
                "driver": "file",
                "filename": self.scratch_disk,
            },
            "backing": self.file,
        })

    def add_dirty_bitmap(self):
        """
        Real backup code get a list of all checkpoints since incremental, and
        merge all of them into the temporary bitmap.
        """
        log.debug("Adding dirty bitmap %s for incremental backup since: %s",
                  self.bitmap, self.incremental)

        self.qmp.execute("block-dirty-bitmap-add", {
            "node": self.file,
            "name": self.bitmap,
            "disabled": True,
        })

        self.qmp.execute("block-dirty-bitmap-merge", {
            "node": self.file,
            "target": self.bitmap,
            "bitmaps": [self.incremental],
        })

    def run_backup_transaction(self):
        log.debug("Running backup transaction")
        actions = []

        if self.incremental:
            # Disbale the previous active dirty bitmap. Changes after this
            # point will be recorded in the new bitmap.
            actions.append({
                "type": "block-dirty-bitmap-disable",
                "data": {
                    "node": self.file,
                    "name": self.incremental,
                }
            })

        if self.checkpoint:
            actions.append({
                "type": "block-dirty-bitmap-add",
                "data": {
                    "name": self.checkpoint,
                    "node": self.file,
                    "persistent": True,
                }
            })

        actions.append({
            "type": "blockdev-backup",
            "data": {
                "device": self.file,
                "job-id": self.job,
                "sync": "none",
                "target": self.node,
            }
        })

        self.qmp.execute("transaction", {"actions": actions})

    def add_to_nbd_server(self):
        log.debug("Adding node %s and bitmap %s to nbd server",
                  self.node, self.bitmap)
        arguments = {
            "type": "nbd",
            "id": self.export,
            "node-name": self.node,
            "name": self.export,
            "allocation-depth": True,
        }
        if self.bitmap:
            arguments["bitmaps"] = [self.bitmap]
        self.qmp.execute("block-export-add", arguments)

    def remove_from_nbd_server(self):
        log.debug("Removing export %s from nbd server", self.export)
        self.qmp.execute("block-export-del", {
            "id": self.export,
            "mode": "hard",
        })

    def stop_nbd_server(self):
        log.debug("Stopping nbd server")
        self.qmp.execute("nbd-server-stop")

    def cancel_block_job(self):
        log.debug("Cancelling block job %s", self.job)
        self.qmp.execute("block-job-cancel", {"device": self.job})

    def remove_backup_node(self):
        log.debug("Removing backup node %s", self.node)
        self.qmp.execute("blockdev-del", {"node-name": self.node})

    def remove_dirty_bitmap(self):
        log.debug("Removing dirty bitmap %s from node %s",
                  self.bitmap, self.file)
        self.qmp.execute("block-dirty-bitmap-remove", {
            "node": self.file,
            "name": self.bitmap,
        })


# Backup data helpers.


def copy_disk(nbd_url, backup_disk):
    log.info("Backing up data extents from %s to %s", nbd_url, backup_disk)
    backup_url = urlparse(nbd_url)

    with nbd.open(backup_url) as src_client, \
            qemu_nbd.open(backup_disk, "qcow2") as dst_client:
        nbdutil.copy(src_client, dst_client)


def copy_dirty(nbd_url, backup_disk):
    log.info("Backing up dirty extents from %s to %s", nbd_url, backup_disk)
    backup_url = urlparse(nbd_url)

    with nbd.open(backup_url, dirty=True) as src_client, \
            qemu_nbd.open(backup_disk, "qcow2") as dst_client:

        buf = bytearray(4 * 1024**2)

        offset = 0
        for ext in nbdutil.extents(src_client, dirty=True):
            if ext.dirty:
                todo = ext.length
                while todo:
                    step = min(todo, len(buf))
                    view = memoryview(buf)[:step]
                    src_client.readinto(offset, view)
                    dst_client.write(offset, view)
                    offset += step
                    todo -= step
            else:
                offset += ext.length



# ovirt-imageio
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


class Closed:

    def __getattr__(self, name):
        """
        Behave like closed io.FileIO.
        """
        raise ValueError("Operation on closed backend")


CLOSED = Closed()



# ovirt-imageio
# Copyright (C) 2015-2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from . import configloader


class daemon:

    # Interval in seconds for checking termination conditions.
    poll_interval = 1.0

    # Maximum number of connections for same /image/ticket-id URL. Using more
    # connections improves throughput of a single image transfer. When
    # transferring images concurrently, using more connections per transfer may
    # decrease throughput.
    max_connections = 8

    # Number of seconds before unauthorized connections are
    # disconnected.
    auth_timeout = 15

    # Number of seconds before inactive connections are disconnected.
    # Clients with special needs can request a larger timeout when
    # creating an image transfer.
    inactivity_timeout = 60

    # Daemon run directory. Runtime stuff like socket or profile information
    # will be stored in this directory.
    # This is configurable only for development purposes and is not expected to
    # be changed by the user, therefore it's not documented in example
    # configuration.
    run_dir = "/run/ovirt-imageio"

    # The daemon is started under root be able to access files owned by root
    # and bind to privileged ports. Once started, privileges are dropped.
    # In the tests we want to run all the tests under the root and need an
    # option how to switch it off.
    # This is configurable only for development purposes and is not expected to
    # be changed by the user, therefore it's not documented in example
    # configuration.
    drop_privileges = True

    # Username under which imageio daemon will be run.
    # This is configurable only for development purposes and is not expected to
    # be changed by the user, therefore it's not documented in example
    # configuration.
    user_name = "ovirtimg"

    # Group under which imageio daemon will be run.
    # This is configurable only for development purposes and is not expected to
    # be changed by the user, therefore it's not documented in example
    # configuration.
    group_name = "ovirtimg"


class tls:

    # Enable TLS. Note that without TLS transfer tickets and image data are
    # transferred in clear text.
    # TODO: update the documentation with instructions, how to create
    # certificate files and link it in config file.
    enable = True

    # Private key file, must not be empty.
    key_file = ""

    # Certificate file, must not be empty.
    cert_file = ""

    # CA certificate file. Empty value is valid, meaning use the host trusted
    # CAs.
    ca_file = ""

    # Enable TLSv1.1, for legacy user applications that do not support
    # TLSv1.2.
    enable_tls1_1 = False


class backend_file:

    # Buffer size in bytes for reading and writing using the file backend. The
    # default value seems to give optimal throughput with both low end and high
    # end storage, using iSCSI and FC. Larger values may increase throughput
    # slightly, but may also decrease it significantly.
    # TODO: Tested with single writer, needs testing with multiple readers.
    buffer_size = 8 * 1024**2


class backend_http:

    # CA certificate file to be used with HTTP backend. Empty value is valid,
    # meaning uses CA file configured in TLS section.
    # This option has to be used when the daemon serving as a proxy use
    # different CA than daemon serving storage backend. In most of the cases
    # these CAs are the same and one would use same value as configured in
    # tls.ca_file option.
    ca_file = ""

    # Buffer size in bytes for handling proxy requests. The default value was
    # copied form the file backend.
    # TODO: Needs testing with multiple readers and writers.
    buffer_size = 8 * 1024**2


class backend_nbd:

    # Buffer size in bytes when reading and writing to the nbd backend. The
    # default value was copied form the file backend.
    # TODO: Needs testing with multiple readers and writers.
    buffer_size = 8 * 1024**2


class remote:

    # Remote service interface. Use "::" to listen on any interface on both
    # IPv4 and IPv6. To listen only on IPv4, use "0.0.0.0".
    host = "::"

    # Remote service port. Changing this value require change in the
    # firewall rules on the host, and changing this value in engine
    # configuration.
    port = 54322


class local:

    # Enable local service.
    enable = True

    # Local service unix socket for accessing images locally.
    socket = "\u0000/org/ovirt/imageio"


class control:

    # Transport be used to communicate with control service socket.
    # Can be either "tcp" or "unix". If "unix" is used, communication will
    # be done over UNIX socket which path is specified in "socket" option.
    # In case of TCP transport, you must specify the port using "port" option.
    # Preferred transport is unix as has better security - only users in
    # ovirtimg group can read/write into the socket.
    transport = "unix"

    # Control service socket path. This socket is used to control the
    # daemon and must be accessible only to the program controlling the
    # daemon.
    socket = "/run/ovirt-imageio/sock"

    # Control service port when run over TCP. Changing this value require
    # changing this value in engine configuration.
    # The default value is not set. If you want to use TCP transport, you
    # must specify port.
    port = -1

    # Determines if IPv4 address should be preferred when the address of
    # control service is resolved. This option allows compatibility with java
    # clients which may prefer IPv4 address and don't try other addresses on
    # dual stack system.
    prefer_ipv4 = True

    # Number of seconds to wait when removing a ticket. If ticket cannot be
    # removed within this timeout, the request will fail with "409 Conflict",
    # and the user need to retry the request again. A ticket can be removed
    # only when the number of connections using the ticket is zero.
    remove_timeout = 60


class profile:

    # Filename for storing profile data. Profiling requires the "yappi"
    # package. Version 0.93 is recommended for best performance.
    filename = "/run/ovirt-imageio/profile"


# Logger configuration.
# See Python logging documentation for details how to configure loggers.

class loggers:
    keys = "root"


class handlers:
    keys = "logfile"


class formatters:
    keys = "long"


class logger_root:
    level = "INFO"
    handlers = "logfile"
    propagate = 0


class handler_logfile:
    keyword__class = "logging.handlers.RotatingFileHandler"
    args = '("/var/log/ovirt-imageio/daemon.log",)'
    kwargs = '{"maxBytes": 20971520, "backupCount": 10}'
    level = "DEBUG"
    formatter = "long"


class handler_stderr:
    keyword__class = "logging.StreamHandler"
    args = "()"
    level = "DEBUG"
    formatter = "long"


class formatter_long:
    format = ("%(asctime)s %(levelname)-7s (%(threadName)s) [%(name)s] "
              "%(message)s")


class Config:

    def __init__(self):
        # Daemon config.

        self.daemon = daemon()
        self.tls = tls()
        self.backend_file = backend_file()
        self.backend_http = backend_http()
        self.backend_nbd = backend_nbd()
        self.remote = remote()
        self.local = local()
        self.control = control()
        self.profile = profile()

        # Logger config.

        self.loggers = loggers()
        self.handlers = handlers()
        self.formatters = formatters()
        self.logger_root = logger_root()
        self.handler_logfile = handler_logfile()
        self.handler_stderr = handler_stderr()
        self.formatter_long = formatter_long()


def load(files):
    cfg = Config()
    configloader.load(cfg, files)
    return cfg


def to_dict(config):
    return configloader.to_dict(config)



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from .. import cors
from .. import version


class Handler:
    """
    Handle requests for the /info resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    @cors.allow()
    def get(self, req, resp):
        resp.send_json({"version": version.string})



# ovirt-imageio
# Copyright (C) 2015-2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import errno
import logging
import os
import stat

from contextlib import closing

from .. import errors
from .. import extent
from .. import ioutil
from .. import util

from . common import CLOSED

log = logging.getLogger("backends.file")


def open(url, mode="r", sparse=False, dirty=False, max_connections=8,
         **options):
    """
    Open a file backend.

    Arguments:
        url (url): parsed file url of underlying file.
        mode: (str): "r" for readonly, "r+" for read write.
        sparse (bool): deallocate space when zeroing if possible.
        dirty (bool): ignored, file backend does not support dirty extents.
        max_connections (int): maximum number of connections per backend
            allowed on this server. Limit backends's max_readers and
            max_writers.
        **options: ignored, file backend does not have any options.
    """
    fio = util.open(url.path, mode, direct=True)
    try:
        fio.name = url.path
        mode = os.fstat(fio.fileno()).st_mode
        backend = BlockBackend if stat.S_ISBLK(mode) else FileBackend
        return backend(fio, sparse=sparse, max_connections=max_connections)
    except:  # noqa: E722
        fio.close()
        raise


class Backend:
    """
    Base class for file backends.
    """

    def __init__(self, fio, sparse=False, max_connections=8):
        """
        Initizlie an I/O backend.

        Arguments:
            fio (io.FileIO): underlying file object.
            sparse (bool): deallocate space when zeroing if possible.
        """
        log.debug("Open path=%r mode=%r sparse=%r max_connections=%r",
                  fio.name, fio.mode, sparse, max_connections)
        self._fio = fio
        self._sparse = sparse
        self._dirty = False
        self._max_connections = max_connections

    @property
    def max_readers(self):
        return self._max_connections

    # io.FileIO interface

    def readinto(self, buf):
        return self._fio.readinto(buf)

    def write(self, buf):
        self._dirty = True
        if (not self._aligned(self.tell()) or len(buf) < self._block_size):
            # The slow path.
            return self._write_unaligned(buf)
        else:
            # The fast path.
            if self._aligned(len(buf)):
                return self._fio.write(buf)
            else:
                count = util.round_down(len(buf), self._block_size)
                with memoryview(buf)[:count] as view:
                    return self._fio.write(view)

    def tell(self):
        return self._fio.tell()

    def seek(self, pos, how=os.SEEK_SET):
        return self._fio.seek(pos, how)

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        try:
            self.close()
        except Exception:
            # Do not hide the original error.
            if t is None:
                raise
            log.exception("Error closing")

    def close(self):
        if self._fio is not CLOSED:
            log.debug("Close path=%r dirty=%r",
                      self._fio.name, self._dirty)
            try:
                self._fio.close()
            finally:
                self._fio = CLOSED

    # Backend interface.

    def zero(self, count):
        """
        Zero up to count bytes starting at current file offset.

        If this backend is opened in sparse mode, the operation will deallocate
        space. Otherwise the operation allocates new space.
        """
        self._dirty = True
        start = self.tell()
        if (not self._aligned(start) or count < self._block_size):
            # The slow path.
            count = min(count, self._block_size - start % self._block_size)
            return self._write_unaligned(b"\0" * count)
        else:
            # The fast path.
            count = util.round_down(count, self._block_size)
            if self._sparse:
                return self._zero_sparse(count)
            else:
                return self._zero(count)

    def flush(self):
        os.fsync(self._fio.fileno())
        self._dirty = False

    @property
    def block_size(self):
        return self._block_size

    def extents(self, context="zero"):
        if context != "zero":
            raise errors.UnsupportedOperation(
                "Backend {} does not support {} extents"
                .format(self.name, context))

        # TODO: Use qemu-img map to get extents.
        yield extent.ZeroExtent(0, self.size(), False, False)

    # Debugging interface

    def readable(self):
        return self._fio.readable()

    def writable(self):
        return self._fio.writable()

    @property
    def dirty(self):
        return self._dirty

    @property
    def sparse(self):
        return self._sparse

    @property
    def name(self):
        return "file"

    def size(self):
        old_pos = self._fio.tell()
        self._fio.seek(0, os.SEEK_END)
        result = self._fio.tell()
        self._fio.seek(old_pos, os.SEEK_SET)
        return result

    # Private

    def _aligned(self, n):
        """
        Return True if number n is aligned to block size.
        """
        return not (n & (self._block_size - 1))

    def _write_unaligned(self, buf):
        """
        Write up to block_size bytes from buf into the current block.

        If position is not aligned to block size, writes only up to end of
        current block.

        Perform a read-modify-write on the current block:
        1. Read the current block
        2. copy bytes from buf into the block
        3. write the block back to storage.

        Returns:
            Number of bytes written
        """
        start = self.tell()
        offset = start % self._block_size
        count = min(len(buf), self._block_size - offset)

        log.debug("Unaligned write start=%s offset=%s count=%s",
                  start, offset, count)

        block = util.aligned_buffer(self._block_size)
        with closing(block):
            # 1. Read available bytes in current block.
            self.seek(start - offset)
            self.readinto(block)

            # 2. Write new bytes into buffer.
            block[offset:offset + count] = buf[:count]

            # 3. Write block back to storage. This aligns the file to block
            # size by padding zeros if needed.
            # TODO: When writing to file system, block size may be wrong, so we
            # need to take care of short writes.
            self.seek(start - offset)
            self._fio.write(block)

            # 4. Update position.
            self.seek(start + count)

        return count

    def _clone(self):
        mode = self._fio.mode.replace("b", "")
        fio = util.open(self._fio.name, mode=mode, direct=True)
        try:
            return self.__class__(
                fio,
                sparse=self._sparse,
                max_connections=self._max_connections,
                block_size=self._block_size)
        except:  # noqa: E722
            fio.close()
            raise


class BlockBackend(Backend):
    """
    Block device backend.
    """

    def __init__(self, fio, sparse=False, max_connections=8, block_size=512):
        """
        Initialize a BlockBackend.

        Arguments:
            fio (io.FileIO): underlying file object.
            sparse (bool): deallocate space when zeroing if possible.
            max_connections (int): maximum number of connections per backend
                allowed on this server. Limit backends's max_readers and
                max_writers.
            block_size (int): If set, use the specified block size. Otherwise
                the value is detected automatically.
        """
        super().__init__(fio, sparse=sparse, max_connections=max_connections)
        # May be set to False if the first call to fallocate() reveal that it
        # is not supported.
        self._can_fallocate = True
        # TODO: get block size using ioctl(BLKSSZGET).
        self._block_size = block_size

    def clone(self):
        """
        Return a new backend sharing the same block device.
        """
        backend = self._clone()
        backend._can_fallocate = self._can_fallocate
        return backend

    @property
    def max_writers(self):
        return self._max_connections

    def _zero(self, count):
        """
        Zero count bytes at current file position, allocating space.
        """
        offset = self.tell()

        # First try fallocate(). It works also for block devices since kernel
        # 4.9. We prefer it since it also invalidates the page cache, avoiding
        # reading stale data.
        if self._can_fallocate:
            mode = ioutil.FALLOC_FL_ZERO_RANGE
            try:
                ioutil.fallocate(self._fio.fileno(), mode, offset, count)
            except EnvironmentError as e:
                # On RHEL 7.5 (kenerl 3.10.0) this will fail with ENODEV.
                if e.errno not in (errno.EOPNOTSUPP, errno.ENODEV):
                    raise
                # fallocate() is not supported - do not try again.
                log.debug("fallocate(mode=%r) is not supported, zeroing "
                          "using BLKZEROOUT",
                          mode)
                self._can_fallocate = False
            else:
                self.seek(offset + count)
                return count

        # If we reach this, this kernel does not support fallocate() for block
        # devices, so we fallback to BLKZEROOUT.
        ioutil.blkzeroout(self._fio.fileno(), offset, count)
        self.seek(offset + count)
        return count

    # Emulate with zero.
    # TODO: With recent kernel we can use fallocate(PUNCH_HOLE). It works for
    # devices whitelisted to zero discarded range.
    _zero_sparse = _zero


class FileBackend(Backend):
    """
    Regular file backend.
    """

    def __init__(self, fio, sparse=False, max_connections=8, block_size=None):
        """
        Initialize a FileBackend.

        Arguments:
            fio (io.FileIO): underlying file object.
            sparse (bool): deallocate space when zeroing if possible.
            max_connections (int): maximum number of connections per backend
                allowed on this server. Limit backends's max_readers.
            block_size (int): If set, use the specified block size. Otherwise
                the value is detected automatically.
        """
        super().__init__(fio, sparse=sparse, max_connections=max_connections)
        # These will be set to False if the first call to fallocate() reveal
        # that it is not supported on the current file system.
        self._can_zero_range = True
        self._can_punch_hole = True
        self._can_fallocate = True
        self._block_size = block_size or self._detect_block_size()

    def clone(self):
        """
        Return a new backend sharing the same file.
        """
        backend = self._clone()
        backend._can_zero_range = self._can_zero_range
        backend._can_punch_hole = self._can_punch_hole
        backend._can_fallocate = self._can_fallocate
        return backend

    @property
    def max_writers(self):
        # Zeroing and trimming qcow2 format grows the file and assumes a single
        # writer. User that wants best performance should use the nbd backend.
        return 1

    def _detect_block_size(self):
        """
        Detect the unserlying storage block size by checking the minimal block
        size that works for direct I/O.

        There are 2 cases when we cannot detect the block size:
        - Reading from unallocated file in local XFS or Gluster over XFS. This
          should not happen now since qemu-img create always allocate the first
          block to mitigate this issue.
        - NFS, since O_DIRECT s not passed to the server

        When we cannot detect the block size we fallback to 4096.
        """
        for block_size in (1, 512, 4096):
            log.debug("Trying block size %s", block_size)
            buf = util.aligned_buffer(block_size)
            with closing(buf):
                self.seek(0)
                try:
                    self.readinto(buf)
                except EnvironmentError as e:
                    if e.errno != errno.EINVAL:
                        raise
                    continue

            self.seek(0)

            if block_size == 1:
                log.debug("Cannot detect block size - using 4096")
                block_size = 4096
            else:
                log.debug("Detected block size %s", block_size)

            return block_size

        raise RuntimeError(
            "Cannot use direct I/O with {}".format(self._fio.path))

    def _zero(self, count):
        """
        Zero count bytes at current file position, allocating space.
        """
        offset = self.tell()

        # First try the modern way. If this works, we can zero a range using
        # single call. Unfortunately, this does not work with NFS 4.2.
        if self._can_zero_range:
            mode = ioutil.FALLOC_FL_ZERO_RANGE
            if self._fallocate(mode, offset, count):
                self.seek(offset + count)
                return count
            else:
                log.debug("Cannot zero range")
                self._can_zero_range = False

        # Next try to punch a hole and then allocate the range. This hack is
        # used by qemu since 2015.
        # See https://github.com/qemu/qemu/commit/1cdc3239f1bb
        if self._can_punch_hole and self._can_fallocate:
            mode = ioutil.FALLOC_FL_PUNCH_HOLE | ioutil.FALLOC_FL_KEEP_SIZE
            if self._fallocate(mode, offset, count):
                if self._fallocate(0, offset, count):
                    self.seek(offset + count)
                    return count
                else:
                    log.debug("Cannot fallocate range")
                    self._can_fallocate = False
            else:
                log.debug("Cannot punch hole")
                self._can_punch_hole = False

        # If we are writing after the end of the file, we can allocate.
        if self._can_fallocate:
            size = os.fstat(self._fio.fileno()).st_size
            if offset >= size:
                if self._fallocate(0, offset, count):
                    self.seek(offset + count)
                    return count
                else:
                    log.debug("Cannot fallocate range")
                    self._can_fallocate = False

        # We have to write zeros manually.
        self._write_zeros(count)
        return count

    def _zero_sparse(self, count):
        """
        Zero count bytes at current file position, punching a hole.
        """
        # First try to punch a hole.
        if self._can_punch_hole:
            offset = self.tell()

            # Extend file size if needed.
            size = os.fstat(self._fio.fileno()).st_size
            if offset + count > size:
                self._truncate(offset + count)

                # If we zero the end of the file punching a hole is not needed.
                if size == offset:
                    self.seek(offset + count)
                    return count

            # And punch a hole.
            mode = ioutil.FALLOC_FL_PUNCH_HOLE | ioutil.FALLOC_FL_KEEP_SIZE
            if self._fallocate(mode, offset, count):
                self.seek(offset + count)
                return count
            else:
                log.debug("Cannot punch hole")
                self._can_punch_hole = False

        # We have to write zeros manually.
        self._write_zeros(count)
        return count

    def _fallocate(self, mode, offset, count):
        """
        Try to fallocate, returning True if the attempt was successful, or
        False if this mode is not supported. Any other error is raised.
        """
        try:
            ioutil.fallocate(self._fio.fileno(), mode, offset, count)
            return True
        except EnvironmentError as e:
            if e.errno != errno.EOPNOTSUPP:
                raise
            return False

    def _truncate(self, size):
        self._fio.truncate(size)

    def _write_zeros(self, count):
        """
        Write zeros manually.
        """
        buf_size = min(count, 1024**2)
        with util.aligned_buffer(buf_size) as buf, memoryview(buf) as view:
            while count:
                count -= self.write(view[:count])



# ovirt-imageio
# Copyright (C) 2021 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import logging
import uuid

import pytest

from ovirt_imageio import admin
from ovirt_imageio._internal import config
from ovirt_imageio._internal import server

from . import testutil

log = logging.getLogger("test")


@pytest.fixture(scope="module", params=["daemon", "proxy"])
def srv(request):
    path = "test/conf/{}.conf".format(request.param)
    cfg = config.load(path)
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


@pytest.mark.parametrize("trans", ["unix", "tcp"])
def test_client_error(trans):
    ticket = testutil.create_ticket(
        ops=["read"], sparse=False, dirty=False, transfer_id="123")

    class cfg:
        class control:
            transport = trans
            socket = "/no/such/sock"
            port = 64 * 1024

    log.debug("Accessing stopped server")
    with pytest.raises(admin.ClientError) as e:
        with admin.Client(cfg, timeout=0.5) as c:
            c.add_ticket(ticket)
    log.debug("Error: %s", e.value)


def test_get_ticket_missing(srv):
    with admin.Client(srv.config) as c:
        ticket_id = str(uuid.uuid4())

        with pytest.raises(admin.ServerError) as e:
            log.debug("Getting ticket %s", ticket_id)
            c.get_ticket(ticket_id)

        log.debug("Error: %s", e.value)
        assert e.value.code == 404
        assert ticket_id in e.value.message


def test_add_ticket(srv):
    ticket = testutil.create_ticket(
        ops=["read"], sparse=False, dirty=False, transfer_id="123")

    with admin.Client(srv.config) as c:
        log.debug("Adding ticket %s", ticket)
        c.add_ticket(ticket)
        info = c.get_ticket(ticket["uuid"])
        log.debug("Got ticket %s", info)
        assert info["uuid"] == ticket["uuid"]


def test_get_ticket(srv, fake_time):
    ticket = testutil.create_ticket(
        ops=["read"], sparse=False, dirty=False, transfer_id="123")

    with admin.Client(srv.config) as c:
        log.debug("Adding ticket %s", ticket)
        c.add_ticket(ticket)
        info = c.get_ticket(ticket["uuid"])
        log.debug("Got ticket %s", info)

        assert info == {
            # Ticket attributes
            "ops": ticket["ops"],
            "size": ticket["size"],
            "sparse": ticket["sparse"],
            "dirty": ticket["dirty"],
            "inactivity_timeout": ticket["inactivity_timeout"],
            "timeout": ticket["timeout"],
            "url": ticket["url"],
            "uuid": ticket["uuid"],
            "transfer_id": ticket["transfer_id"],
            # Ticket server status.
            "active": False,
            "canceled": False,
            "connections": 0,
            "expires": 300,
            "idle_time": 0,
            "transferred": 0
        }


def test_mod_ticket(srv, fake_time):
    ticket = testutil.create_ticket(
        ops=["read"], sparse=False, dirty=False, transfer_id="123")

    with admin.Client(srv.config) as c:
        log.debug("Adding ticket %s", ticket)
        c.add_ticket(ticket)
        info = c.get_ticket(ticket["uuid"])
        assert info["expires"] == 300

        # Extend ticket.
        c.mod_ticket(ticket["uuid"], {"timeout": 600})
        info = c.get_ticket(ticket["uuid"])
        assert info["expires"] == 600

        # Expire ticket.
        c.mod_ticket(ticket["uuid"], {"timeout": 0})
        info = c.get_ticket(ticket["uuid"])
        assert info["expires"] == 0

        # Try to change read only value
        with pytest.raises(admin.ServerError):
            c.mod_ticket(ticket["uuid"], {"url": "file:///other"})

        # Value must not change.
        info = c.get_ticket(ticket["uuid"])
        assert info["url"] == ticket["url"]


def test_del_ticket(srv):
    ticket = testutil.create_ticket(
        ops=["read"], sparse=False, dirty=False, transfer_id="123")

    with admin.Client(srv.config) as c:
        log.debug("Adding ticket %s", ticket)
        c.add_ticket(ticket)

        # Make sure ticket exists before removing it.
        c.get_ticket(ticket["uuid"])

        log.debug("Removing ticket %s", ticket["uuid"])
        c.del_ticket(ticket["uuid"])

        with pytest.raises(admin.ServerError) as e:
            c.get_ticket(ticket["uuid"])

        log.debug("Error: %s", e.value)
        assert e.value.code == 404


def test_del_ticket_missing(srv):
    with admin.Client(srv.config) as c:
        ticket_id = str(uuid.uuid4())
        # Shoudld succeed.
        c.del_ticket(ticket_id)



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Compatibility module for vdsm.

vdsm uses ovirt_imageio_common.directio.Receive. We plan to remove the
dependency, but we cannot fix old vdsm versions.

Changes to this module must be tested with /usr/libexec/vdsm/kvm2ovirt.
"""

import urllib.parse as urllib_parse

from . import ops
from . import stats
from . import util
from . backends import file


class Receive(ops.Write):
    """
    ops.Write accepts a backend instead of path, and buffer instead of
    buffersize. Provide a backward compatible version for existing users (e.g.
    vdsm).
    """

    def __init__(self, path, src, size=None, offset=0, flush=True,
                 buffersize=1024**2, clock=stats.NullClock()):
        url = urllib_parse.urlparse("file:" + path)
        self._dst = file.open(url, "r+")
        buf = util.aligned_buffer(buffersize)
        super().__init__(self._dst, src, buf, size=size, offset=offset,
                         flush=flush, clock=clock)

    def close(self):
        if self._dst:
            try:
                self._dst.close()
            finally:
                self._dst = None
            try:
                self._buf.close()
            finally:
                self._buf = None
            super().close()
